from sym.sdk.user import UserIdentity


def persist_user_identity(
    *, email: str, service: str, service_id: str, user_id: str
) -> UserIdentity:
    """This function will create or update a :class:`~sym.sdk.user.UserIdentity`
    for a :class:`~sym.sdk.user.User` identified by the provided email.

    If no :class:`~sym.sdk.user.User` matching the provided email exists,
    one will be created.

    Args:
        email: The email of the Sym :class:`~sym.sdk.user.User` to persist a
            :class:`~sym.sdk.user.UserIdentity` for.
        service: The name of the external system providing the identity (e.g. "slack").
        service_id: The ID of the external system providing the identity (e.g. "T123ABC")
            for a Slack Workspace.
        user_id: The :class:`~sym.sdk.user.User`'s identifier in the external system
            (e.g. U12345 for a Slack User ID).

    Returns:
        The :class:`~sym.sdk.user.UserIdentity` that was created or updated.

    Raises:
        :class:`~sym.sdk.exceptions.identity.CouldNotSaveError`: if the identity
            failed to save for any reason.
    """



"""Classes for representing Strategies and Integrations, including fully custom Strategies."""

__all__ = ["AccessStrategy", "Integration"]

from .access_strategy import AccessStrategy
from .integration import Integration



import pytest

from sym.sdk.exceptions import PagerDutyError
from sym.sdk.integrations.pagerduty import is_on_call, users_on_call


class TestPagerdutyInterface:
    def test_is_on_call(self):
        user = {"username": "jon.doe@simi.org"}
        assert is_on_call(user) is None

    def test_users_on_call(self):
        assert users_on_call() is None

    def test_error_code(self):
        with pytest.raises(PagerDutyError) as error_info:
            raise PagerDutyError(name="UNKNOWN", message="foobar")

        assert error_info.value.error_code == "PagerDuty:UNKNOWN"
        assert error_info.value.message == "foobar"



"""Helpers for interacting with the AWS IAM API within the Sym SDK."""

from sym.sdk.exceptions.aws import AWSIAMError  # noqa
from sym.sdk.user import User


def is_user_in_group(user: User, *, group_name: str) -> bool:
    """Checks if the provided user is a member of the AWS IAM group specified.

    The unique AWS IAM group name must be given, and the method will check that the group exists and is
    accessible. An exception will be thrown if not.

    Args:
        user: The user to check group membership of.
        group_name: The name of the AWS IAM group.

    Returns:
        True if the user is a member of the specified AWS IAM group, False otherwise.
    """



"""Helpers for interacting with the GitHub API within the Sym SDK."""


from typing import Iterable, List, Literal, Union

from sym.sdk.exceptions import GitHubError  # noqa
from sym.sdk.user import User

GitHubDefaultRolesT = Literal["read", "triage", "write", "maintain", "admin"]
GitHubRoleT = Union[GitHubDefaultRolesT, str]


def get_repo_collaborators(*, repo_name: str, roles: Iterable[GitHubRoleT]) -> List[User]:
    """Get all collaborators of the specified repository with the given role(s), mapped to their
    Sym User objects.

    This is useful to, for example, set the approvers of a request to be users with "admin" access
    to the specified repository.

    Note that ``roles`` are mutually exclusive in GitHub—for example, users with the "admin" role
    technically have write permissions, but not the "write" role. Because of this, "admin" users
    would not be listed if the "write" role were requested. To retrieve both sets of users, you may
    specify multiple roles, e.g., ``{"admin", "write"}``.

    Also note that ``repo_name`` should not include the organization; the organization is derived
    from your Terraform configuration.

    Caution: This method will only return collaborators that have a Sym user corresponding to their
    GitHub Identity. Users can be added and identities managed using the symflow CLI.

    Args:
        repo_name: The name, not including organization, of the repository to get the collaborators
            of
        roles: A set of one or more roles on which to filter collaborators.

    Returns:
        A list of :class:`Users <sym.sdk.user.User>` corresponding to collaborators with the specified role(s)
        of the specified repository.
    """



from sym.sdk.exceptions.sym_exception import ExceptionWithHint, SymException


class IdentityError(SymException):
    """This is the base class for errors that occur when managing identities.

    Args:
        name: The name of the exception (used as the second part of the error code, e.g. COULD_NOT_SAVE)
        message: The exception message to display
    """

    def __init__(self, name: str, message: str, error_type: str = "Identity"):
        super().__init__(error_type=error_type, name=name, message=message)


class CouldNotSaveError(IdentityError):
    """This error is raised in cases where a :class:`~sym.sdk.user.UserIdentity`
    was unable to be saved.
    """

    def __init__(self, message: str):
        super().__init__("COULD_NOT_SAVE", message)


class IdentityNotFound(IdentityError, ExceptionWithHint):
    """This error is raised in cases where a mapping of a Sym :class:`~sym.sdk.user.User` to an
    external identity (i.e., a :class:`~sym.sdk.user.UserIdentity`) could not be found and could not
    be automatically determined from the corresponding service.

    Args:
        email: The email address of the user for which identity lookup could not be performed
        service_id: The unique Implementer-provided identifier of the service for which identity
            lookup was needed
    """

    def __init__(self, email: str, service_id: str):
        super().__init__(
            "NOT_FOUND",
            f"Could not find or determine external identity for {email} in service {service_id}",
        )

    @property
    def hint(self) -> str:
        return "An Implementer can use the symflow CLI to manually add the user's identity"



from sym.sdk.exceptions.sym_exception import SymException


class OktaError(SymException):
    """This is the base class for all OktaErrors.

    Args:
        name: The name of the exception (used as the second part of the error code, e.g. GROUP_NOT_FOUND)
        message: The exception message to display
    """

    def __init__(self, name: str, message: str, error_type: str = "Okta"):
        super().__init__(error_type=error_type, name=name, message=message)



"""Tools for describing Sym Resources."""

import json
import re
from typing import Optional, Union

from .errors import SymSDKError


class InvalidSRNError(SymSDKError):
    """Raised when an invalid :class:`~sym.sdk.resource.SRN` is supplied."""

    def __init__(self, srn: str, hint: str = None):
        self.srn = srn
        super().__init__(
            message=f"Invalid SRN '{srn}'.",
            hint=(
                hint
                or "SRNs must match the following structure: <ORG>:<MODEL>[:<TYPE>]:<SLUG>:<VERSION>[:<IDENTIFIER>], where org, model, slug, and version are required."
            ),
            doc_url="https://docs.symops.com/docs/sym-concepts",
        )


class InvalidSlugError(InvalidSRNError):
    """Raised when a component of a :class:`~sym.sdk.resource.SRN` is an invalid slug."""

    def __init__(self, srn: str, component: str, value: str):
        super().__init__(
            srn,
            f"The {component} must be a valid slug (alphanumeric characters and dashes only). Got: {value}",
        )


class MissingComponentError(InvalidSRNError):
    """Raised when a component of a :class:`~sym.sdk.resource.SRN` is missing."""

    def __init__(self, srn: str, component: str):
        super().__init__(
            srn,
            f"The {component} component is required.",
        )


class InvalidVersionError(InvalidSRNError):
    """Raised when a :class:`~sym.sdk.resource.SRN` has an invalid version."""

    def __init__(self, srn: str, component: str, value: str):
        super().__init__(
            srn,
            f"The {component} must be semver with no tags (e.g. 1.0.0). Got: {value}",
        )


class TrailingSeparatorError(InvalidSRNError):
    """Raised when a :class:`~sym.sdk.resource.SRN` contains a trailing separator."""

    def __init__(self, srn: str):
        super().__init__(srn, "SRNs cannot have a trailing separator.")


class MultipleErrors(InvalidSRNError):
    """Raised when a :class:`~sym.sdk.resource.SRN` has multiple validation errors."""

    def __init__(self, srn: str):
        super().__init__(srn)
        self.errors = []

    def add(self, exc: InvalidSRNError):
        self.errors.append(exc)
        self.hint = "\n- ".join(
            ["There were multiple validation errors."] + [x.hint for x in self.errors]
        )

    def check(self):
        if len(self.errors) == 1:
            raise self.errors[0]
        elif self.errors:
            raise self


class SRN:
    """Sym Resource Name (:class:`~sym.sdk.resource.SRN`) is a unique identifier for a Sym Resource.

    SRNs have the following structure::

        <ORG>:<MODEL>[:<TYPE>]:<SLUG>:<VERSION>[:<IDENTIFIER>]

    Where VERSION is either a semver string, or "latest". And TYPE indicates the type of the model; this is often
    the `type` field of the resource defined in Terraform. For example, the type `slack` for an `integration` resource.

    For example, the :class:`~sym.sdk.resource.SRN` for the v1.0.0 sym:approval
    template is::

        sym:template:approval:1.0.0

    Or the :class:`~sym.sdk.resource.SRN` for a :class:`~sym.sdk.flow.Flow`
    instance (with a UUID as an instance identifier) could be::

        sym:flow:test-flow:0.1.0:d47782bc-88be-44df-9e34-5fae0dbdea22

    Or the :class:`~sym.sdk.resource.SRN` for a Slack integration with a slug "my-integration" is::

        sym:integration:slack:my-integration:latest:d47782bc-88be-44df-9e34-5fae0dbdea22
    """

    SEPARATOR = ":"
    """The default separator for :class:`~sym.sdk.resource.SRN` components."""

    SLUG_PATTERN = re.compile(r"^[a-zA-Z0-9-_]+$")
    """The pattern for validating slug components."""
    VERSION_PATTERN = re.compile(r"^(latest|[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})$")
    """The pattern for validating the version component."""

    _COMPONENTS = {
        "org": (SLUG_PATTERN, InvalidSlugError),
        "model": (SLUG_PATTERN, InvalidSlugError),
        "type": (SLUG_PATTERN, InvalidSlugError),
        "slug": (SLUG_PATTERN, InvalidSlugError),
        "version": (VERSION_PATTERN, InvalidVersionError),
        "identifier": (SLUG_PATTERN, InvalidSlugError),
    }

    _MODELS_WITH_TYPES = {
        "access_strategy",
        "access_target",
        "user",
        "secret",
        "secrets",
        "integration",
        "log_destination",
    }

    @classmethod
    def parse(cls, raw: str) -> "SRN":
        """Parses and validates the given string as an :class:`~sym.sdk.resource.SRN`.

        Args:
            raw: A raw string representing a :class:`~sym.sdk.resource.SRN`.

        Returns:
            A :class:`~sym.sdk.resource.SRN` instance.

        Raises:
            :class:`~sym.sdk.resource.TrailingSeparatorError`:      The string has a trailing separator.
            :class:`~sym.sdk.resource.InvalidSRNError`:             The string is missing components, or at least one component is invalid.
            :class:`~sym.sdk.resource.InvalidSlugError`:            The string has an invalid slug component.
            :class:`~sym.sdk.resource.InvalidVersionError`:         The string has an invalid version component.
        """
        raw = str(raw).lower()

        if raw.endswith(cls.SEPARATOR):
            raise TrailingSeparatorError(raw)

        parts = raw.split(cls.SEPARATOR)

        # SRNs will always have at least 4 parts (org, model, slug, version)
        if len(parts) < 4:
            raise InvalidSRNError(raw)

        model = parts[1]
        if cls.requires_type(model):
            # If this model has a type, enforce that it exists
            if len(parts) < 5:
                raise InvalidSRNError(
                    raw,
                    "This SRN must have at least 5 components: org, model, type, slug, version",
                )

            return cls(
                org=parts[0],
                model=model,
                type=parts[2],
                slug=parts[3],
                version=parts[4],
                identifier=parts[5] if len(parts) > 5 else None,
            )
        else:
            return cls(
                org=parts[0],
                model=model,
                type=None,
                slug=parts[2],
                version=parts[3],
                identifier=parts[4] if len(parts) > 4 else None,
            )

    @classmethod
    def __get_validators__(cls):
        yield cls.parse

    @classmethod
    def __modify_schema__(cls, field_schema: dict):
        """A hook to allow Pydantic to export a JSON Schema for SRN fields.

        https://pydantic-docs.helpmanual.io/usage/types/#custom-data-types
        """
        slug = cls.SLUG_PATTERN.pattern[1:-1]
        version_pattern = cls.VERSION_PATTERN.pattern[1:-1]
        field_schema.update(
            type="string",
            description="A Sym Resource Name (SRN) is a unique identifier for a Sym Resource.",
            pattern=f"{slug}:{slug}(:{slug})?:{slug}:{version_pattern}(:{slug})?",
            examples=["sym:event:approval:1.0.0:requested"],
        )

    def __init__(
        self,
        org: str,
        model: str,
        type: Optional[str],
        slug: str,
        version: str,
        identifier: Optional[str] = None,
    ):
        self._org = org.lower()
        self._model = model.lower()
        self._type = type.lower() if type else type
        self._slug = slug.lower()
        self._version = version
        self._identifier = identifier.lower() if identifier else identifier

        self._validate()

    def __str__(self):
        return self.SEPARATOR.join(
            [x for x in [self._get(k) for k in self._COMPONENTS.keys()] if x is not None]
        )

    def __repr__(self) -> str:
        components = ", ".join(
            [
                f"{k}={v}"
                for (k, v) in [(k, self._get(k)) for k in self._COMPONENTS.keys()]
                if v is not None
            ]
        )
        return f"SRN({components})"

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, (SRN, str)):
            return False

        if isinstance(other, str):
            try:
                other = SRN.parse(other)
            except Exception:
                return False

        return (
            self._org == other._org
            and self._model == other._model
            and self._type == other._type
            and self._slug == other._slug
            and self._version == other._version
            and self._identifier == other._identifier
        )

    def _get(self, name: str):
        return getattr(self, f"_{name}")

    def _validate(self):
        errors = MultipleErrors(str(self))
        for name, (pattern, error_class) in self._COMPONENTS.items():
            component_value = self._get(name)
            if component_value is not None and not pattern.match(component_value):
                errors.add(error_class(str(self), name, component_value))

        resource_has_type = self.requires_type(self._get("model"))

        if resource_has_type and self._get("type") is None:
            errors.add(MissingComponentError(str(self), "type"))

        if not resource_has_type and (type_component := self._get("type")):
            errors.add(
                InvalidSRNError(
                    str(self),
                    f"Resource type {self._get('model')} cannot have a type component. Got: {type_component}",
                )
            )

        errors.check()

    @classmethod
    def requires_type(cls, model: str) -> bool:
        """Returns True if the given model has a type component."""
        return model in cls._MODELS_WITH_TYPES

    def copy(
        self,
        organization: Optional[str] = None,
        model: Optional[str] = None,
        type: Optional[str] = None,
        slug: Optional[str] = None,
        version: Optional[str] = None,
        identifier: Optional[str] = None,
    ):
        """Creates a copy of this :class:`~sym.sdk.resource.SRN`.

        Optionally can create a new :class:`~sym.sdk.resource.SRN` with
        modified components from the current, as specified by the keyword arguments.
        """

        components = [
            organization or self._org,
            model or self._model,
            type or self._type,
            slug or self._slug,
            version or self._version,
        ]
        if identifier:
            components.append(identifier)
        elif self._identifier:
            components.append(self._identifier)

        return self.__class__(*components)

    @property
    def organization(self) -> str:
        """The slug for the organization this :class:`~sym.sdk.resource.SRN`
        belongs to.

        For example, for the sym:approval :class:`~sym.sdk.templates.template.Template`,
        the organization slug is `sym`.
        """

        return self._org

    @property
    def model(self) -> str:
        """The model name for this :class:`~sym.sdk.resource.SRN`.

        For example, for the sym:approval :class:`~sym.sdk.templates.template.Template`,
        the model name is `template`.
        """

        return self._model

    @property
    def type(self) -> Optional[str]:
        """The model type for this :class:`~sym.sdk.resource.SRN`.

        For example, for a Slack integration SRN `sym:integration:slack:my-integration:latest`,
        the type is `slack`.

        If no type is specified, then this property will return None.
        """
        return self._type

    @property
    def slug(self) -> str:
        """This :class:`~sym.sdk.resource.SRN`'s slug.

        For example, for the sym:approval :class:`~sym.sdk.templates.template.Template`, the slug is `approval`.
        """
        return self._slug

    @property
    def version(self) -> str:
        """A semver string representing the version of this :class:`~sym.sdk.resource.SRN`.

        For example, the first version of the sym:approval :class:`~sym.sdk.templates.template.Template`
        is `1.0.0`.
        """

        return self._version

    @property
    def identifier(self) -> Optional[str]:
        """An arbitrary string identifying an instance of the resource.

        This is often a UUID.
        """
        return self._identifier


class SymBaseResource:
    """The base class that all Sym SDK models inherit from."""

    def __str__(self) -> str:
        return json.dumps(self.dict(), indent=2)

    def __repr__(self) -> str:
        return str(self)

    def dict(self):
        """Represent this resource as a dictionary."""
        return {
            k: v.dict() if isinstance(v, SymBaseResource) else v
            for k in dir(self)
            if not k.startswith("_")
            and isinstance((v := getattr(self, k)), (SymBaseResource, str, int, dict, list))
        }

    def __eq__(self, other):
        if not isinstance(other, SymBaseResource):
            return False
        return self.dict() == other.dict()


class SymResource(SymBaseResource):
    """A piece of infrastructure provisioned with
    Sym's `Terraform provider <https://docs.symops.com/docs/terraform-provider>`_.

    For example, a :class:`~sym.sdk.flow.Flow` is a Resource.

    Read more about `Sym Resources <https://docs.symops.com/docs/sym-concepts>`_.
    """

    def __init__(self, srn: Union[SRN, str]):
        self._srn = SRN.parse(str(srn))

    def __getattr__(self, name: str):
        """__getattr__ is called as a last resort if there are no attributes on the
        instance that match the name.

        This override allows attributes of the SRN to be called as attributes directly\
        on the resource without needing to define them.
        """
        if name in {"srn", "_srn"}:
            # Raise if we've failed to find SRN, otherwise we'll infinitely recurse.
            raise AttributeError(f"'{self._class_name}' object has no attribute '{name}'")

        try:
            return getattr(self.srn, name)
        except AttributeError:
            # If `name` doesn't exist in SRN, then the original attribute error will report
            # 'SRN' object has no attribute 'name', but we should report the original object name for clarity.
            raise AttributeError(f"'{self._class_name}' object has no attribute '{name}'")

    def __eq__(self, other):
        if not isinstance(other, SymResource):
            return False
        return self.srn == other.srn

    def __hash__(self):
        return hash(str(self.srn))

    def __repr__(self) -> str:
        return f"{self._class_name}({self.srn})"

    @property
    def _class_name(self) -> str:
        """Extracts the name of a class, making some public-friendly replacements."""
        return self.__class__.__name__.replace("SDK", "")

    @property
    def srn(self) -> SRN:
        """A :class:`~sym.sdk.resource.SRN` object that represents the unique identifier
        for this resource.
        """
        return self._srn

    @property
    def name(self) -> str:
        """An alias for this resource's slug, derived from its :class:`~sym.sdk.resource.SRN`."""
        return self.srn.slug



from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional

from sym.sdk.resource import SRN

from .template import Template


class ApprovalTemplateStep(str, Enum):
    """The :class:`~sym.sdk.templates.approval.ApprovalTemplateStep` enum lists
    the steps in the :class:`~sym.sdk.templates.approval.ApprovalTemplate`, each of which
    can be used in hooks to fire new :class:`Events <sym.sdk.event.Event>`.

    Read more about the `steps of the sym:approval template <https://docs.symops.com/docs/sym-approval#steps>`_.
    """

    PROMPT = "prompt"
    REQUEST = "request"
    REQUEST_FORWARD = "request_forward"
    REQUEST_EXPIRE = "request_expire"
    APPROVE = "approve"
    DENY = "deny"
    ESCALATE = "escalate"
    DEESCALATE = "deescalate"


class ApprovalTemplate(Template, ABC):
    """The :class:`~sym.sdk.templates.approval.ApprovalTemplate` object represents
    a security workflow for access management supported out of the box by Sym.

    Read more about the `sym:approval <https://docs.symops.com/docs/sym-approval>`_
    :class:`~sym.sdk.templates.template.Template`.
    """

    @classmethod
    @abstractmethod
    def prompt(cls, **kwargs):
        """Generates a request to fire an :class:`~sym.sdk.event.Event` of type ``prompt`` to
        pop up a modal for the :class:`~sym.sdk.user.User` to make a request.

        Args:
            **kwargs: Arbitrary additional values to pass through to the fired :class:`~sym.sdk.event.Event`'s payload.
        """

    @classmethod
    @abstractmethod
    def request(cls, *, target_srn: Optional[SRN] = None, duration: Optional[int] = None, **kwargs):
        """Generates a request to fire an :class:`~sym.sdk.event.Event` of type ``request`` to
        submit a request for access to an :class:`~sym.sdk.target.AccessTarget`.

        Args:
            target_srn: The :class:`~sym.sdk.resource.SRN` of the :class:`~sym.sdk.target.AccessTarget` to request access to.
                Required only if firing an :class:`~sym.sdk.event.Event` for a :class:`~sym.sdk.flow.Run` which has not had
                a request submitted. Otherwise, defaults to the current :class:`~sym.sdk.flow.Run`'s requested :class:`~sym.sdk.target.AccessTarget`.
            duration: How long the escalation should last.
                Required only if firing an :class:`~sym.sdk.event.Event` for a :class:`~sym.sdk.flow.Run` which has not had
                a request submitted. Otherwise, defaults to the current :class:`~sym.sdk.flow.Run`'s requested duration.
            **kwargs: Arbitrary additional values to pass through to the fired :class:`~sym.sdk.event.Event`'s payload.
        """

    @classmethod
    @abstractmethod
    def approve(cls, *, target_srn: Optional[SRN] = None, duration: Optional[int] = None, **kwargs):
        """Generates a request to fire an :class:`~sym.sdk.event.Event` of type ``approve`` to
        approve an outstanding request for access to an :class:`~sym.sdk.target.AccessTarget`.

        Args:
            target_srn: The :class:`~sym.sdk.resource.SRN` of the :class:`~sym.sdk.target.AccessTarget` to request access to.
                Required only if firing an :class:`~sym.sdk.event.Event` for a :class:`~sym.sdk.flow.Run` which has not had
                a request submitted. Otherwise, defaults to the current :class:`~sym.sdk.flow.Run`'s requested :class:`~sym.sdk.target.AccessTarget`.
            duration: How long the escalation should last.
                Required only if firing an :class:`~sym.sdk.event.Event` for a :class:`~sym.sdk.flow.Run` which has not had
                a request submitted. Otherwise, defaults to the current :class:`~sym.sdk.flow.Run`'s requested duration.
            **kwargs: Arbitrary additional values to pass through to the fired :class:`~sym.sdk.event.Event`'s payload.
        """

    @classmethod
    @abstractmethod
    def deny(cls, *, target_srn: Optional[SRN] = None, duration: Optional[int] = None, **kwargs):
        """Generates a request to fire an :class:`~sym.sdk.event.Event` of type ``deny`` to
        deny an outstanding request for access to an  :class:`~sym.sdk.target.AccessTarget`.

        Args:
            target_srn: The :class:`~sym.sdk.resource.SRN` of the :class:`~sym.sdk.target.AccessTarget` to request access to.
                Required only if firing an :class:`~sym.sdk.event.Event` for a :class:`~sym.sdk.flow.Run` which has not had
                a request submitted. Otherwise, defaults to the current :class:`~sym.sdk.flow.Run`'s requested :class:`~sym.sdk.target.AccessTarget`.
            duration: How long the escalation should last.
                Required only if firing an :class:`~sym.sdk.event.Event` for a :class:`~sym.sdk.flow.Run` which has not had
                a request submitted. Otherwise, defaults to the current :class:`~sym.sdk.flow.Run`'s requested duration.
            **kwargs: Arbitrary additional values to pass through to the fired :class:`~sym.sdk.event.Event`'s payload.
        """

    @classmethod
    @abstractmethod
    def escalate(
        cls, *, target_srn: Optional[SRN] = None, duration: Optional[int] = None, **kwargs
    ):
        """Generates a request to fire an :class:`~sym.sdk.event.Event` of type ``escalate`` to
        begin escalation of a :class:`~sym.sdk.user.User` for an :class:`~sym.sdk.target.AccessTarget`.

        Args:
            target_srn: The :class:`~sym.sdk.resource.SRN` of the :class:`~sym.sdk.target.AccessTarget` to escalate the :class:`~sym.sdk.user.User` for.
                Required only if firing an :class:`~sym.sdk.event.Event` for a :class:`~sym.sdk.flow.Run` which has not had
                a request submitted. Otherwise, defaults to the current :class:`~sym.sdk.flow.Run`'s requested :class:`~sym.sdk.target.AccessTarget`.
            duration: How long the escalation should last.
                Required only if firing an :class:`~sym.sdk.event.Event` for a :class:`~sym.sdk.flow.Run` which has not had
                a request submitted. Otherwise, defaults to the current :class:`~sym.sdk.flow.Run`'s requested duration.
            **kwargs: Arbitrary additional values to pass through to the fired :class:`~sym.sdk.event.Event`'s payload.
        """

    @classmethod
    @abstractmethod
    def deescalate(cls, *, target_srn: Optional[SRN] = None, **kwargs):
        """Generates a request to fire an :class:`~sym.sdk.event.Event` of type ``deescalate`` to
        begin deescalation of a :class:`~sym.sdk.user.User` for an :class:`~sym.sdk.target.AccessTarget`.

        Args:
            target_srn: The :class:`~sym.sdk.resource.SRN` of the :class:`~sym.sdk.target.AccessTarget` to deescalate the :class:`~sym.sdk.user.User` for.
                Required only if firing an :class:`~sym.sdk.event.Event` for a :class:`~sym.sdk.flow.Run` which has not had
                a request submitted. Otherwise, defaults to the current :class:`~sym.sdk.flow.Run`'s requested :class:`~sym.sdk.target.AccessTarget`.
            **kwargs: Arbitrary additional values to pass through to the fired :class:`~sym.sdk.event.Event`'s payload.
        """



"""Helpers for interacting with the Okta API within the Sym SDK."""


from typing import List, Optional

from sym.sdk.exceptions import OktaError  # noqa
from sym.sdk.user import User


def is_user_in_group(user: User, *, group_id: str) -> bool:
    """Checks if the provided user is a member of the Okta group specified.

    The Okta group's ID must be given, and the method will check that the group exists and is
    accessible. An exception will be thrown if not.

    Args:
        user: The user to check group membership of.
        group_id: The ID of the Okta group.

    Returns:
        True if the user is a member of the specified Okta group, False otherwise.
    """


def users_in_group(*, group_id: str) -> List[User]:
    """Get all users from the specified Okta group.

    The Okta group's ID must be given, and the method will check that the group exists and is
    accessible. An exception will be thrown if not.

    Args:
        group_id: The ID of the Okta group.
    """


def get_user_info(user: User) -> dict:
    """Returns Okta User Info for the Sym user.

    The user info follows the data format of Okta's Users API, details here:
    https://developer.okta.com/docs/reference/api/users/#response-example-10

    Args:
        user: The Sym user to request Okta info for

    Returns:
        A dict of user info
    """


def list_groups(
    *,
    name_filter: Optional[str] = None,
    filter: Optional[str] = None,
    expand: Optional[List[str]] = None,
    only_native_groups: bool = True,
) -> List[dict]:
    """Returns a list of Okta Group Info.

    The arguments match the arguments in Okta's List Groups API,
    details `here <https://developer.okta.com/docs/reference/api/groups/#request-parameters-3>`__

    Args:
        name_filter: Finds a group whose name starts with ``name_filter``, case-insensitive.
        filter: A filter expression for groups. See `here <https://developer.okta.com/docs/reference/core-okta-api/#filter>`__
                for syntax and details.
                Note: The expressions for filter queries must use double quotes:

                .. code-block:: python

                    filter='id eq "00g12345"'
        expand: If specified, additional metadata will be included in the response.
                Possible values are: ``stats`` and/or ``app``.
        only_native_groups: If true, only groups of source type "Native Okta" (i.e. the only group source type modifiable by Sym)
                will be included in the results. See
                `here <https://help.okta.com/en-us/Content/Topics/users-groups-profiles/usgp-group-types.htm>`__
                for more information on Okta group source types.

    Returns:
        A list of dictionaries, with each dictionary representing a Group in Okta's
        Group structure. See
        `here <https://developer.okta.com/docs/reference/api/groups/#group-attributes>`__
        for details.
    """



import pytest

from sym.sdk.templates import ApprovalTemplate, Template


class TestTemplates:
    def test_approval(self):
        with pytest.raises(TypeError, match="Can't instantiate abstract class ApprovalTemplate"):
            ApprovalTemplate("sym:template:approval:1.0.0")

    def test_abstract_base(self):
        with pytest.raises(TypeError):
            Template()



import pytest

from sym.sdk.target import AccessTarget


class TestTarget:
    def test_access_target_abstract(self):
        with pytest.raises(TypeError, match="Can't instantiate abstract class AccessTarget"):
            AccessTarget()  # type: ignore



"""Workflow templates that can be declaratively provisioned."""

__all__ = [
    "ApprovalTemplate",
    "ApprovalTemplateStep",
    "Template",
    "TemplateStep",
    "get_step_output",
]

from .approval import ApprovalTemplate, ApprovalTemplateStep
from .step import get_step_output
from .template import Template, TemplateStep



"""An instance of something to request access to."""

from abc import ABC, abstractmethod
from typing import Optional

from sym.sdk.resource import SymResource


class AccessTarget(SymResource, ABC):
    """The :class:`~sym.sdk.target.AccessTarget` class represents an instance
    of something a :class:`~sym.sdk.user.User` can request access to.

    Read more about `Targets <https://docs.symops.com/docs/tf-target>`_.
    """

    @property
    @abstractmethod
    def label(self) -> Optional[str]:
        """The display name for this :class:`~sym.sdk.target.AccessTarget`."""

    @property
    @abstractmethod
    def type(self) -> str:
        """The type of resource to be accessed (e.g., "okta_group", "aws_iam_group")."""

    @property
    @abstractmethod
    def settings(self) -> dict:
        """The dictionary of settings from :class:`~sym.sdk.target.AccessTarget`'s
        definition in Terraform.
        """



"""A parameterized instance of a :class:`~sym.sdk.templates.template.Template`."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict

from sym.sdk.resource import SymResource
from sym.sdk.user import User

if TYPE_CHECKING:
    from sym.sdk.event import Channel
    from sym.sdk.strategies.integration import Integration


class Environment(SymResource, ABC):
    """The :class:`~sym.sdk.flow.Environment` class represents a container
    that shares configuration values across :class:`~sym.sdk.flow.Flow` instances.

    Read more about `Environments <https://docs.symops.com/docs/tf-environment>`_.
    """

    @property
    @abstractmethod
    def integrations(self) -> Dict[str, "Integration"]:
        """A dictionary containing all :class:`Integrations <sym.sdk.strategies.integration.Integration>` that may be used in the implementation for a :class:`~sym.sdk.flow.Flow`, as defined in Terraform.


        For example, if Slack and AWS Lambda :class:`Integrations <sym.sdk.strategies.integration.Integration>` are specified in the :class:`~sym.sdk.flow.Environment`, the ``integrations`` dictionary would look like this::

            {
                "slack": <sym.sdk.strategies.integration.Integration>,
                "aws_lambda": <sym.sdk.strategies.integration.Integration>
            }

        """


class Flow(SymResource, ABC):
    """The :class:`~sym.sdk.flow.Flow` class contains a particular parameterization
    of a :class:`~sym.sdk.templates.template.Template`.

    Read more about `Flows <https://docs.symops.com/docs/tf-flow>`_.
    """

    @property
    @abstractmethod
    def fields(self) -> Dict[str, Any]:
        """A dictionary with the field structure defined for the :class:`~sym.sdk.flow.Flow`
        in Terraform.
        """

    @property
    @abstractmethod
    def vars(self) -> Dict[str, str]:
        """A dict containing user-supplied values from the :class:`~sym.sdk.flow.Flow`'s definition in Terraform.

        This dict might contain, for example, your team's PagerDuty schedule ID.
        """

    @property
    @abstractmethod
    def environment(self) -> Environment:
        """A reference to the :class:`~sym.sdk.flow.Environment` attached to this :class:`~sym.sdk.flow.Flow`."""


class Run(SymResource, ABC):
    """A :class:`~sym.sdk.flow.Run` represents an instance of a :class:`~sym.sdk.flow.Flow`
    in progress.

    For example, each new access request using the `sym:template:approval:1.0`
    :class:`~sym.sdk.templates.template.Template` would generate a new :class:`~sym.sdk.flow.Run`
    with data pertaining to that specific access request.
    """

    @property
    @abstractmethod
    def actors(self) -> Dict[str, User]:
        """A dict mapping :class:`~sym.sdk.event.Event` names to the :class:`~sym.sdk.user.User`
        that created each :class:`~sym.sdk.event.Event`. There will be one entry for each
        :class:`~sym.sdk.event.Event` in the current :class:`~sym.sdk.flow.Run`.

        For example, with a sym:approval :class:`~sym.sdk.flow.Flow`, after the "approve"
        :class:`~sym.sdk.event.Event` is received, the actors may look like this::

            {
                "prompt": <User A>,
                "request": <User A>,
                "approve": <User B>
            }
        """

    @property
    @abstractmethod
    def contexts(self) -> Dict[str, Dict[str, Any]]:
        """A dict mapping :class:`~sym.sdk.event.Event` names to the context attached to
        each :class:`~sym.sdk.event.Event`. There will be one entry for each
        :class:`~sym.sdk.event.Event` in the current :class:`~sym.sdk.flow.Run`.

        For example, with a sym:approval :class:`~sym.sdk.flow.Flow`, after the "approve"
        :class:`~sym.sdk.event.Event` is received, the contexts may look like this::

            {
                "request": {
                    "additional_id": "12345"
                }
            }
        """

    @property
    @abstractmethod
    def source_channel(self) -> "Channel":
        """A :class:`~sym.sdk.event.Channel` object, which contains information about the where
        the :class:`~sym.sdk.flow.Run`'s first :class:`~sym.sdk.event.Event` came from.
        """



"""Helpers for interacting with the AWS SSO API within the Sym SDK."""
from typing import List, Optional

from sym.sdk.exceptions.aws import AWSSSOError  # noqa
from sym.sdk.user import User


def is_user_in_group(user: User, *, group_name: str) -> bool:
    """Checks if the provided :class:`~sym.sdk.user.User` is a member of the AWS SSO group specified.

    The AWS SSO group's display name must be given, and the method will check that the group exists and is
    accessible. An exception will be raised if not.

    Args:
        user: The :class:`~sym.sdk.user.User` to check group membership of.
        group_name: The display name of the AWS SSO group.

    Returns:
        True if the :class:`~sym.sdk.user.User` is a member of the specified AWS SSO group, False otherwise.

    Raises:
        :class:`~sym.sdk.exceptions.aws.AWSSSOError`: If no AWS SSO group with the given name exists.
    """


def list_groups() -> List[dict]:
    """Lists all the groups in the identity store of the AWS SSO account connected by an AWS SSO Sym Integration.
    See `here <https://docs.symops.com/docs/aws-sso-access#add-a-aws-sso-integration>`__ for instructions on how
    to create an AWS SSO Integration.

    Returns:
        A list of dictionaries, with each dictionary representing a Group in AWS SSO's Group structure.
        See `here <https://docs.aws.amazon.com/singlesignon/latest/IdentityStoreAPIReference/API_Group.html>`__
        for details.
    """


def users_in_group(group_name: str) -> List[User]:
    """Get all users from the specified AWS SSO group.

    *Warning: This method may only be used for groups that have 20 or fewer members.*

    The AWS SSO group's display name must be given, and the method will check that the group exists and is
    accessible. An exception will be raised if not.

    Args:
        group_name: The display name of the AWS SSO group.

    Returns:
        A list of :class:`~sym.sdk.user.User`

    Raises:
        :class:`~sym.sdk.exceptions.aws.AWSSSOError`: If no AWS SSO group with the given name exists or if the group has greater than 20 members.
    """


def list_accounts(organizational_unit_id: Optional[str] = None) -> List[dict]:
    """Returns a list of AWS accounts within the AWS SSO Organization that the AWS SSO Sym Integration is connected to,
    optionally filtered to the specified Organizational Unit.

    The behavior of this function changes based on whether the ``organizational_unit_id`` parameter is specified, and if
    so what value is provided.

    - If unspecified or set to ``None``, this function will return all AWS accounts within your Organization.
    - If a parent root ID (beginning with ``r-``) is specified, this function will return all AWS accounts *not* mapped
      to any Organizational Unit.
    - If an Organizational Unit ID (any other value) is specified, this function will return all AWS accounts mapped to
      that Organizational Unit *only*; accounts in child OUs of the specified OU will *not* be included.

    Args:
        organizational_unit_id: The ID of the parent root or Organizational Unit to filter by, or ``None`` to include
            all AWS accounts.

    Returns:
        A list of dictionaries, with each dictionary representing an AWS account. The dictionaries follow the AWS
        Account object structure; see
        `here <https://docs.aws.amazon.com/organizations/latest/APIReference/API_Account.html>`__ for details.

    Raises:
        :class:`~sym.sdk.exceptions.aws.AWSSSOError`: If the specified ``organizational_unit_id`` is invalid or does
            not exist, if and only if ``organizational_unit_id`` is specified.
    """



from abc import abstractmethod


class SymException(Exception):
    """This is the base class for all exceptions raised by the Sym Runtime.

    Args:
        error_type: The class of exception (used as the first part of the error_code, e.g. AuthenticationError)
        name: The name of the exception (used as the second part of the error_code, e.g. INVALID_JWT)
        message: The exception message to display
    """

    def __init__(self, error_type: str, name: str, message: str):
        self.message = message
        self.error_code = f"{error_type}:{name}"
        super().__init__(self.message)

    def to_dict(self):
        return {
            "error": True,
            "message": self.message,
            "code": self.error_code,
        }


class ExceptionWithHint(SymException):
    """A trait that can be mixed into Exception classes that want to have messages formatted with a
    hint.

    Overrides the ``__str__`` method to provide a formatted output string.

    i.e. the output will look like
    > This is my main exception message
    >
    > *Hint:* This is the exception hint

    Caution: This class should not be used for matching exceptions in ``except`` statements!
    """

    @property
    @abstractmethod
    def hint(self) -> str:
        """The string to be displayed in the Hint section"""
        ...

    def __str__(self):
        return f"{self.message}\n\n*Hint:* {self.hint}"



from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional

from sym.sdk.resource import SymResource


class TemplateStep(str, Enum):
    """The :class:`~sym.sdk.templates.template.TemplateStep` class represents general steps which may
    be part of any :class:`~sym.sdk.templates.template.Template`.
    """

    IGNORE = "ignore"
    """Fire an :class:`~sym.sdk.event.Event` of type ``ignore`` to ullify the incoming
    :class:`~sym.sdk.event.Event` and send a message to the acting :class:`~sym.sdk.user.User`
    """


class Template(SymResource, ABC):
    """The :class:`~sym.sdk.templates.template.Template` object represents a
    common security workflow supported out of the box by Sym.

    Read more about `Templates <https://docs.symops.com/docs/available-templates>`_.
    """

    @classmethod
    @abstractmethod
    def ignore(cls, *, message: Optional[str] = None, **kwargs):
        """Generates a request to fire an :class:`~sym.sdk.event.Event` of type ``ignore`` to
        nullify the incoming :class:`~sym.sdk.event.Event` and send a message to the acting :class:`~sym.sdk.user.User`.

        Args:
            message: The message to send to the :class:`~sym.sdk.user.User` who triggered the original :class:`~sym.sdk.event.Event`.
            **kwargs: Arbitrary additional values to pass through to the fired :class:`~sym.sdk.event.Event`'s payload.
        """



"""Common errors that can be thrown."""
from enum import Enum
from typing import Optional


class SymSDKError(Exception):
    """The base exception class used by Sym's SDK."""

    # To ensure all user-facing errors are informative and actionable,
    # all SymSDKErrors are required to have an error message, a hint as to
    # the next step a user should take, and a relevant link to Sym documentation.
    def __init__(self, message: str, hint: str, doc_url: str, **kwargs):
        super().__init__(message, **kwargs)
        self.message = message
        self.hint = hint
        self.doc_url = doc_url

    def __str__(self):
        return f"{self.message}\n\nHint: {self.hint}\n\nFor more details, see: {self.doc_url}."

    def format_slack(self) -> str:
        """Returns a string representation of this error formatted for display
        in Slack-compatible markdown.
        """
        return f"*{self.message}*\n\n*Hint:* {self.hint}\n\nFor more details, see: {self.doc_url}."


class SymIntegrationError(SymSDKError):
    """Raised when there is an error in an Integration.

    Each exception includes an error code that you can use to reach out to support.
    """

    default_message = "An error occurred."
    default_hint = "Please contact support."
    default_docs_url = "https://docs.symops.com/docs/support"

    def __init__(
        self,
        message: str = "",
        hint: Optional[str] = None,
        doc_url: Optional[str] = None,
        *,
        error_code: Optional[Enum] = None,
        params: Optional[dict] = None,
    ):
        """
        Args:
            error_code: A valid Sym error code
            hint: A helpful suggestion to the User
            params: A dict of relevant values
        """
        self.error_code = error_code
        self.params = params

        super().__init__(
            message=message or self.default_message,
            hint=hint or self.default_hint,
            doc_url=doc_url or self.default_docs_url,
        )

    def _format(self, s: str) -> str:
        try:
            return s.format_map(self.params or {})
        except (KeyError, AttributeError):
            return s

    def format_slack(self) -> str:
        return self._format(super().format_slack())

    def __str__(self):
        return self._format(super().__str__())

    def __repr__(self):
        return self._format(super().__repr__())



from typing import Optional

from sym.sdk.templates import ApprovalTemplateStep


def get_step_output(step: Optional[ApprovalTemplateStep] = None) -> dict:
    """
    Returns the output returned by the specified step (or the current step if no step is specified).

    For Lambda flows, this method can be used in `after_escalate` and `after_deescalate` hooks
    to retrieve responses from the lambda.

    For example::

        @hook
        def after_escalate(evt):
            escalate_output = get_step_output()
            print(escalate_output["body"])

    For :class:`custom Strategies <sym.sdk.strategies.access_strategy.AccessStrategy>`, this method
    can be used in :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.deescalate` to retrieve
    the output from the corresponding escalation by specifying the
    :class:`ApprovalTemplateStep.ESCALATE <sym.sdk.templates.approval.ApprovalTemplateStep>` step::

        def deescalate(self, target_id, event):
            escalate_output = get_step_output(ApprovalTemplateStep.ESCALATE)
            escalation_id = escalate_output["id"]
            ...

    Args:
        step: The step for which to retrieve output. If ``None`` or omitted, returns the output for
            the current step.
    """



from typing import Any, Dict, Optional

import pytest

from sym.sdk import Environment, Flow
from sym.sdk.resource import (
    SRN,
    InvalidSlugError,
    InvalidSRNError,
    MultipleErrors,
    SymResource,
    TrailingSeparatorError,
)


class MockSymResource(SymResource):
    """This is a test class for SymResource behavior."""


class SDKFlow(Flow, SymResource):
    """A stubbed Flow class for testing purposes"""

    @property
    def environment(self) -> Environment:
        pass

    @property
    def fields(self) -> Dict[str, Any]:
        return dict()

    @property
    def vars(self) -> Dict[str, str]:
        return dict()


class TestResource:
    def test_sym_resource_fails_on_malformed_srn(self):
        self._test_bad("foo", InvalidSRNError)
        self._test_bad("foo:bar", InvalidSRNError)
        self._test_bad("foo:bar:baz", InvalidSRNError)
        self._test_bad("foo:bar:baz:boz", InvalidSRNError)
        self._test_bad("foo:bar:baz:lates:", TrailingSeparatorError, match="trailing separator.")
        self._test_bad("foo:bar:baz:1.0:", InvalidSRNError)
        self._test_bad("foo:bar:baz:1.0.0::", InvalidSRNError)
        self._test_bad("foo:bar:baz:latest:1.0.0:", TrailingSeparatorError)
        self._test_bad("foo:bar:baz:1.3000.0:something", InvalidSRNError)
        self._test_bad("foo:bar:baz:something", InvalidSRNError)
        self._test_bad("foo:bar:baz:latestsomething", InvalidSRNError)
        self._test_bad("foo:bar:baz:latest:", TrailingSeparatorError)
        self._test_bad("foo!foobar:bar:baz:latest:foo", InvalidSlugError, match="org")
        self._test_bad("sym:flow:something::", InvalidSRNError)
        self._test_bad("foo!foobar:bar:baz:1000.0.2000:foo", MultipleErrors, match="version")
        self._test_bad("sym:integration:missing-type:1.0.0", InvalidSRNError)
        self._test_bad("sym:integration:missing-type:1.0.0:1234-5667", InvalidSRNError)
        self._test_bad("sym:integration::empty-type:1.0.0", InvalidSlugError, match="type")

    def _test_bad(self, srn, exc, match: str = None):
        if match:
            with pytest.raises(exc, match=match):
                SRN.parse(srn)
        else:
            with pytest.raises(exc):
                SRN.parse(srn)

    def test_sym_srn_succeeds_on_valid_srn(self):
        self._test_good(
            "sym:foo-bar:12345-11233:0.1.0:stuff",
            "sym",
            "foo-bar",
            None,
            "12345-11233",
            "0.1.0",
            "stuff",
        )
        self._test_good(
            "foo:bar:baz:1.300.0:something", "foo", "bar", None, "baz", "1.300.0", "something"
        )
        self._test_good("foo:bar:baz:latest", "foo", "bar", None, "baz", "latest", None)
        self._test_good("foo_foo:bar:baz:latest", "foo_foo", "bar", None, "baz", "latest", None)
        self._test_good(
            "sym:template:approval:1.0.0", "sym", "template", None, "approval", "1.0.0", None
        )
        self._test_good(
            "sym:template:approval:1.0.0:e97af6b3-0249-4855-971f-4e1dd188773a",
            "sym",
            "template",
            None,
            "approval",
            "1.0.0",
            "e97af6b3-0249-4855-971f-4e1dd188773a",
        )
        self._test_good(
            "sym:template:approval:1.0.0",
            "sym",
            "template",
            None,
            "approval",
            "1.0.0",
            None,
        )
        self._test_good(
            "sym:integration:slack:my-integration:latest",
            "sym",
            "integration",
            "slack",
            "my-integration",
            "latest",
            None,
        )

    def _test_good(
        self,
        raw,
        org: str,
        model: str,
        model_type: Optional[str],
        slug: str,
        version: str,
        identifier: Optional[str],
    ):
        srn = SRN.parse(raw)
        assert srn.organization == org
        assert srn.model == model
        assert srn.type == model_type
        assert srn.slug == slug
        assert srn.version == version
        assert srn.identifier == identifier

    def test_srn_copy_should_succeed_without_identifier(self):
        srn_string = "foo:bar:baz:1.0.0"

        srn = SRN.parse(srn_string)

        assert str(srn.copy(version="latest")) == "foo:bar:baz:latest"
        assert str(srn.copy(organization="myorg")) == "myorg:bar:baz:1.0.0"

    def test_srn_str_should_produce_an_identical_srn(self):
        text = "sym:template:approval:1.0.0"
        srn = SRN.parse(text)

        srn_str = str(srn)
        srn2 = SRN.parse(srn_str)

        assert srn == srn2
        assert str(srn) == str(srn2)
        assert text == srn_str

    def test_sym_resource_srn_getattr_fallback(self):
        srn = SRN.parse("test:mock:slug:latest:12345")
        resource = MockSymResource(srn=srn)

        assert resource.srn == srn
        assert resource.name == srn.slug
        assert resource.organization == srn.organization
        assert resource.identifier == srn.identifier

    def test_sym_resource_srn_getattr_errors(self):
        srn = SRN.parse("test:mock:slug:latest:12345")
        resource = MockSymResource(srn=srn)

        # Ensure a normal missing attribute errors properly.
        with pytest.raises(AttributeError, match="no attribute 'nope'"):
            resource.nope

        # Delete the attributes so the __getattr__ override should fall back,
        # but we want to ensure it doesn't infinitely recurse.
        delattr(resource, "_srn")
        with pytest.raises(AttributeError, match="no attribute 'srn"):
            resource.srn

    def test_srn_type(self):
        for model in SRN._MODELS_WITH_TYPES:
            # Check that type is parsed properly
            srn = SRN.parse(f"sym:{model}:type:slug:latest")
            assert srn.type == "type"

            # Check that type is enforced
            with pytest.raises(InvalidSRNError):
                SRN.parse(f"sym:{model}:slug:latest")

            with pytest.raises(InvalidSRNError):
                SRN.parse(f"sym:{model}:slug:latest:identifier")

    def test_srn_copy_with_types(self):
        integration_srn = SRN.parse("test:integration:slack:slug:latest:12345")
        assert integration_srn.type == "slack"

        # Cannot copy from a typed resource to an untyped resource.
        with pytest.raises(InvalidSRNError, match="cannot have a type component"):
            integration_srn.copy(model="flow")

        flow_srn = SRN.parse("test:flow:slug:latest:12345")
        assert str(flow_srn) == "test:flow:slug:latest:12345"
        assert flow_srn.type is None

        # The type must be passed in if copying from an untyped resource to a typed resource.
        with pytest.raises(InvalidSRNError, match="type"):
            flow_srn.copy(model="integration")

        copied_integration_srn = flow_srn.copy(model="integration", type="slack")
        assert str(copied_integration_srn) == "test:integration:slack:slug:latest:12345"
        assert copied_integration_srn.type == "slack"

    def test_srn_init(self):
        SRN("test", "flow", None, "slug", "latest", "12345")

        with pytest.raises(InvalidSRNError, match="type"):
            SRN("test", "integration", None, "slug", "1.0.0")

        with pytest.raises(InvalidSRNError, match="cannot have a type"):
            SRN("test", "flow", "invalid-type", "slug", "latest")

    def test_requires_type(self):
        for model in SRN._MODELS_WITH_TYPES:
            assert SRN.requires_type(model)

        assert not SRN.requires_type("flow")

    def test_eq(self):
        srn1 = SRN.parse("test:flow:slug:latest:12345")
        srn2 = SRN(
            org="test", model="flow", type=None, slug="SLUG", version="latest", identifier="12345"
        )

        # SRNs are equal despite casing
        assert srn1 == srn2
        assert srn1 == "test:flow:slug:latest:12345" == srn2

        # SRNs can equal strings
        assert srn1 == "test:flow:slug:latest:12345"

        # Missing identifier
        assert not srn1 == "test:flow:slug:latest"

        # Slugs not equal
        assert not srn1 == "test:flow:other-slug:latest:12345"

        # Slugs not equal
        assert not srn1 == SRN.parse("test:flow:other-slug:latest:12345")

        # Equality doesn't explode, even if other object is not a SRN
        assert not srn1 == 1
        assert not srn1 == "something"

    def test_get_attr_error(self):
        flow = SDKFlow("test:flow:other-slug:latest:12345")

        # Sanity test, SRN exists, and the getattr convenience method works for
        # attributes that exist in SRN.
        assert flow.srn
        assert flow.srn.organization

        # Validate that error reports "ImplementedFlow", not "SRN"
        # If the attribute is not found in flow AND not in SRN
        with pytest.raises(AttributeError) as e:
            flow.__getitem__("run")

        assert str(e.value) == "'Flow' object has no attribute '__getitem__'"

        with pytest.raises(AttributeError) as e:
            flow.__getattr__("srn")

        assert str(e.value) == "'Flow' object has no attribute 'srn'"

    def test_repr(self):
        flow = SDKFlow("test:flow:other-slug:latest:12345")
        assert flow.__repr__() == "Flow(test:flow:other-slug:latest:12345)"



"""Helpers for interacting with the PagerDuty API."""

from enum import Enum
from typing import Any, Dict, List, Optional

from sym.sdk.exceptions import PagerDutyError  # noqa
from sym.sdk.user import User


class PagerDutyStatus(str, Enum):
    """Represents possible incident statuses in PagerDuty."""

    TRIGGERED = "triggered"
    ACKNOWLEDGED = "acknowledged"
    RESOLVED = "resolved"


class PagerDutyUrgency(str, Enum):
    """Represents possible incident urgencies in PagerDuty."""

    LOW = "low"
    HIGH = "high"


def is_on_call(
    user: User,
    *,
    escalation_policy_name: Optional[str] = None,
    escalation_policy_id: Optional[str] = None,
    schedule_name: Optional[str] = None,
    schedule_id: Optional[str] = None,
) -> bool:
    """Checks if the provided user is currently on-call according to PagerDuty.

    If a name or ID is provided for either escalation policy or schedule, checks if the user is
    on-call for specified escalation policy or schedule.

    If no name or ID is provided for either escalation policy or schedule, checks if the user is
    on-call for ANY escalation policy or schedule.

    Args:
        escalation_policy_name: The name of a specific Escalation Policy to check.
        escalation_policy_id: The ID of a specific Escalation Policy to check.
        schedule_name: The name of a specific Schedule to check.
        schedule_id: The ID of a specific Schedule to check.
    """


def users_on_call(
    *,
    escalation_policy_name: Optional[str] = None,
    escalation_policy_id: Optional[str] = None,
    schedule_name: Optional[str] = None,
    schedule_id: Optional[str] = None,
) -> List[User]:
    """Get all on-call users for the specified escalation policy or schedule from PagerDuty.

    Escalation policy or schedule can be specified by name or ID. If none are provided, returns
    on-call users for ALL escalation policies + schedules.

    Args:
        escalation_policy_name: The name of a specific Escalation Policy to check.
        escalation_policy_id: The ID of a specific Escalation Policy to check.
        schedule_name: The name of a specific Schedule to check.
        schedule_id: The ID of a specific Schedule to check.
    """


def users_for_schedule(
    *,
    schedule_name: Optional[str] = None,
    schedule_id: Optional[str] = None,
) -> List[User]:
    """Get all users for the specified schedule from PagerDuty.

    Schedule can be specified by name or ID.

    Args:
        schedule_name: The name of a specific Schedule to check.
        schedule_id: The ID of a specific Schedule to check.
    """


def get_incidents(
    *,
    assignees: Optional[List[User]] = None,
    service_ids: Optional[List[str]] = None,
    statuses: Optional[List[PagerDutyStatus]] = None,
    urgencies: Optional[List[PagerDutyUrgency]] = None,
    since: Optional[str] = None,
    until: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """Retrieve incidents from PagerDuty based on provided filter params.

    Args:
        assignees: When provided, only incidents assigned to one of these users will be returned.
        service_ids: When provided, only incidents in one of these services will be returned.
        statuses: When provided, only incidents with these statuses will be returned. Defaults to
            ``["TRIGGERED", "ACKNOWLEDGED"]``. Note: If ``assignees`` is provided, ``statuses`` cannot
            contain ``RESOLVED``.
        urgencies: When provided, only incidents with these urgencies will be returned. Defaults to
            ``["HIGH"]``.
        since: When provided, only incidents on or after ``since`` will be returned. Defaults to 24
            hours ago if not provided. Must be an ISO-8601 datetime.
        until: When provided, only incidents before ``until`` will be returned. Defaults to 1
            month after ``since`` if not provided. Must be an ISO-8601 datetime.

    Returns:
        A list of dictionaries, with each dictionary representing an incident in PagerDuty's
        ``incident`` structure. See
        `here <https://developer.pagerduty.com/api-reference/9d0b4b12e36f9-list-incidents#Responses>`_
        for details.
    """


def has_incident(
    *,
    assignees: Optional[List[User]] = None,
    service_ids: Optional[List[str]] = None,
    statuses: Optional[List[PagerDutyStatus]] = None,
    urgencies: Optional[List[PagerDutyUrgency]] = None,
    since: Optional[str] = None,
    until: Optional[str] = None,
) -> bool:
    """Return whether there exists an incident that matches the filters provided in the params.

    Args:
        assignees: When provided, only incidents assigned to one of these users will be returned.
        service_ids: When provided, only incidents in one of these services will be returned.
        statuses: When provided, only incidents with these statuses will be returned. Defaults to
            ``["TRIGGERED", "ACKNOWLEDGED"]``. Note: If ``assignees`` is provided, ``statuses`` cannot
            contain ``RESOLVED``.
        urgencies: When provided, only incidents with these urgencies will be returned. Defaults to
            ``["HIGH"]``.
        since: When provided, only incidents on or after ``since`` will be returned. Defaults to 24
            hours ago if not provided. Must be an ISO-8601 datetime.
        until: When provided, only incidents before ``until`` will be returned. Defaults to 1
            month after ``since`` if not provided. Must be an ISO-8601 datetime.

    Returns:
        A boolean representing whether there is an incident matching the given filter parameters.
    """



"""Helpers for interacting with a Slack workspace."""

from typing import Optional, Sequence, Union

from sym.sdk.exceptions.slack import SlackError  # noqa
from sym.sdk.request_destination import (
    RequestDestination,
    RequestDestinationFallback,
    SlackChannelID,
    SlackChannelName,
    SlackUser,
    SlackUserGroup,
)
from sym.sdk.user import User


def user(
    identifier: Union[str, User], allow_self: bool = False, timeout: Optional[int] = None
) -> SlackUser:
    """A reference to a Slack user.

    Args:
        identifier: The unique Slack user ID, email, or @mention for a Slack user, or a Sym :class:`~sym.sdk.user.User`
            instance. (e.g. "U12345", "jane\@symops.io", "@Jane Austen").
        allow_self: A boolean indicating whether or not the requester may approve this Request.
        timeout: An integer representing the number of seconds for the request to remain active for this user. After
            timeout elapses, the request message will expire and can no longer be interacted with.
    """  # noqa: W605 (The "\@" is the correct way avoid implicit mailto links in Sphinx/rST)


def mention(identifier: Union[str, User]) -> str:
    """Returns a string that mentions a Slack user given their identifier.

    Users can be specified with a Slack user ID, email,
    a string version of an @mention. (e.g. "U12345", "jane\@symops.io", "@Jane Austen", or a
    Sym :class:`~sym.sdk.user.User` instance).
    """  # noqa: W605 (The "\@" is the correct way avoid implicit mailto links in Sphinx/rST)


def channel(
    name: str, allow_self: bool = False, timeout: Optional[int] = None
) -> Union[SlackChannelID, SlackChannelName]:
    """A reference to a Slack channel.

    Args:
        name: The unique channel name or channel ID. (e.g. '#sym-requests' or 'C12345').
        allow_self: A boolean indicating whether or not the requester may approve this Request.
        timeout: An integer representing the number of seconds for the request to remain active for this user. After
            timeout elapses, the request message will expire and can no longer be interacted with.
    """


def group(
    users: Sequence[Union[str, User]], allow_self: bool = False, timeout: Optional[int] = None
) -> SlackUserGroup:
    """A reference to a Slack group DM.

    Args:
        users: A list of Slack Users to include in the group. Users can be specified with a Slack user ID, email,
            a string version of an @mention. (e.g. "U12345", "jane\@symops.io", "@Jane Austen", or a
            Sym :class:`~sym.sdk.user.User` instance).
        timeout: An integer representing the number of seconds for the request to remain active for this user. After
            timeout elapses, the request message will expire and can no longer be interacted with.
    """  # noqa: W605 (The "\@" is the correct way avoid implicit mailto links in Sphinx/rST)


def fallback(
    *destinations: RequestDestination,
    continue_on_delivery_failure: Optional[bool] = True,
    continue_on_timeout: Optional[bool] = False,
) -> RequestDestinationFallback:
    """Returns a :class:`~sym.sdk.request_destination.RequestDestinationFallback` that contains the specified
    :class:`~sym.sdk.request_destination.RequestDestination` objects, with `continue_on_delivery_failure=True`.

    If used as the return value of a `get_approvers` reducer, when a Sym Request is made,
    each :class:`~sym.sdk.request_destination.RequestDestination` will be attempted in sequence until one succeeds.

    For example::

        def get_approvers(event):
            return slack.fallback(slack.channel("#missing"), slack.user("@david"))

    Args:
        destinations: At least 2 :class:`~sym.sdk.request_destination.RequestDestination` objects.
        continue_on_delivery_failure: A boolean representing whether to deliver to the next destination if a failure
            occurs while delivering to the current destination.
        continue_on_timeout: A boolean representing whether to deliver to the next destination if the Request at the
            current destination times out.
    """


def send_message(destination: Union[User, RequestDestination], message: str) -> None:
    """Sends a simple message to a destination in Slack. Accepts either a :class:`~sym.sdk.user.User`
    or a :class:`~sym.sdk.request_destination.RequestDestination`, which may represent a user, group, or channel
    in Slack.

    For example::

        # To send to #general:
        slack.send_message(slack.channel("#general"), "Hello, world!")

        # To DM a specific user:
        slack.send_message(slack.user("me@symops.io"), "It works!")

        # To DM the user who triggered an event:
        slack.send_message(event.user, "You did a thing!")

    Args:
        destination: Where the message should go.
        message: The text contents of the message to send.
    """


def get_user_info(user: User) -> dict:
    """Get information about a Slack user.

    Refer to Slack's users.info API documentation for the details on the response format:
    https://api.slack.com/methods/users.info

    Args:
        user: A Sym :class:`~sym.sdk.user.User` instance to get info about.
    """


def list_users(active_only: bool = True) -> list:
    """Returns a list of all users in the workspace.

    Refer to Slack's users.info API documentation for the details on the response format:
    https://api.slack.com/methods/users.list

    Args:
        active_only: A boolean indicating whether or not to exclude deactivated and deleted users.
    """



"""Provides a namespaced key-value store.

Note:
    This module is currently in private beta. Please reach out if you'd like to test it.
"""


from typing import Any


def fetch(path: str):
    """Fetches a namespaced value from the key-value store.

    Namespaces can be nested with dot notation.
    e.g. fetch("users.titles.alice")

    Note:
        This function is currently in private beta. Please reach out if you'd like to test it.

    Args:
        path (str): A string in dot notation representing a nested namespace.

    Raises:
        KeyError: If a non-namespace object has already been set at a non-leaf node of ``path``.
    """


def put(path: str, value: Any):
    """Puts a namespaced value into the key-value store.

    Namespaces can be nested with dot notation.
    e.g. put("users.titles.alice", "manager")

    Note:
        This function is currently in private beta. Please reach out if you'd like to test it.

    Args:
        path (str): A string in dot notation representing a nested namespace.
        value: Any serializable value to store.

    Raises:
        KeyError: If a non-namespace object has already been set at a non-leaf node of ``path``.
    """



from pydantic import BaseModel


class FieldOption(BaseModel):
    """A Pydantic model representing an option to display in a dynamic Slack drop-down menu.

    FieldOptions are returned by Prefetch Reducers to dynamically
    generate a list of options for Prompt Fields with `prefetch = true`.
    """

    value: str
    """The actual value to pass through when selected."""

    label: str
    """A short label to display in the Slack drop-down menu.

    The label will be used to filter the options to display when a requester types into the Slack input box.
    """



"""Triggers for the various steps of a :class:`~sym.sdk.events.Flow`."""

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, Optional

from sym.sdk.flow import Flow, Run
from sym.sdk.resource import SRN, SymResource
from sym.sdk.templates import Template
from sym.sdk.user import User


class Payload(SymResource):
    """The :class:`~sym.sdk.events.Payload` object contains the data of the
    :class:`~sym.sdk.events.Event`.
    """

    @property
    @abstractmethod
    def timestamp(self) -> datetime:
        """A datetime object indicating when the :class:`~sym.sdk.events.Event` was created."""

    @property
    @abstractmethod
    def fields(self) -> Dict[str, Any]:
        """A dict containing the values submitted by the user who created
        the :class:`~sym.sdk.events.Event`
        """

    @property
    @abstractmethod
    def srn(self) -> SRN:
        """The :class:`~sym.sdk.resource.SRN` of the :class:`~sym.sdk.events.Event` instance."""

    @property
    @abstractmethod
    def user(self) -> User:
        """The :class:`~sym.sdk.user.User` who triggered the :class:`~sym.sdk.events.Event`."""


class Channel(SymResource):
    """The :class:`~sym.sdk.event.Channel` object contains information about the channel
    from which an event was sent.
    """

    @property
    @abstractmethod
    def identifier(self) -> Optional[str]:
        """The identifier of the channel (e.g., ``#general`` for a Slack channel).

        Note that, if the event originated from a Slack *shortcut* (as opposed to
        a slash command), identifier will be ``None``. This is because shortcuts
        are global and not linked to any specific channel.
        """

    @property
    @abstractmethod
    def type(self) -> str:
        """The channel type (e.g., ``slack`` or ``sym``)."""


class EventMeta(SymResource, ABC):
    """Contains metadata about an :class:`~sym.sdk.event.Event` instance."""


class Event(SymResource, ABC):
    """The :class:`~sym.sdk.events.Event` class contains information on an event which has been
    received by Sym, routed to a :class:`~sym.sdk.flow.Run` of a :class:`~sym.sdk.flow.Flow`, and is
    triggering specific user-defined Handlers.

    Each Handler will be invoked with a single argument, which is an instance of this class.
    This :class:`~sym.sdk.events.Event` instance will describe the current execution state,
    and can be used to dynamically alter the behavior and control flow of
    Templates.

    Read more about `Handlers <https://docs.symops.com/docs/workflow-handlers>`_.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the :class:`~sym.sdk.event.Event`."""

    @property
    @abstractmethod
    def payload(self) -> Payload:
        """A :class:`~sym.sdk.event.Payload` object, which contains the primary data
        of the :class:`~sym.sdk.event.Event`.
        """

    @property
    @abstractmethod
    def meta(self) -> EventMeta:
        """An :class:`~sym.sdk.event.EventMeta` object, which contains metadata
        about the :class:`~sym.sdk.event.Event` instance.
        """

    @property
    @abstractmethod
    def template(self) -> Template:
        """A :class:`~sym.sdk.templates.template.Template` object, indicating which
        :class:`~sym.sdk.templates.template.Template` the current :class:`~sym.sdk.flow.Flow` inherits from.
        """

    @property
    @abstractmethod
    def flow(self) -> Flow:
        """A :class:`~sym.sdk.flow.Flow` object, indicating the :class:`~sym.sdk.flow.Flow` that
        the current :class:`~sym.sdk.flow.Run` is an instance of.
        """

    @property
    @abstractmethod
    def run(self) -> Run:
        """A :class:`~sym.sdk.flow.Run` object, indicating the current
        :class:`~sym.sdk.flow.Run`.
        """

    @property
    @abstractmethod
    def channel(self) -> Channel:
        """A :class:`~sym.sdk.event.Channel` object indicating the channel the
        current :class:`~sym.sdk.event.Event` instance is coming from.
        """

    @property
    def user(self) -> User:
        """The :class:`~sym.sdk.user.User` who triggered the :class:`~sym.sdk.event.Event`."""
        return self.payload.user

    @abstractmethod
    def get_actor(self, event_name: str) -> Optional[User]:
        """Retrieve the :class:`~sym.sdk.user.User` who triggered a specific
        :class:`~sym.sdk.event.Event`.

        For example, for a :class:`~sym.sdk.flow.Run` using the
        :class:`~sym.sdk.templates.approval.ApprovalTemplate`, to get the approver::

            approver = event.get_actor("approve")

        For a list of event names, see the relevant template's Enum (e.g.
        :class:`~sym.sdk.templates.approval.ApprovalTemplateStep`).

        Note that, for the :class:`~sym.sdk.templates.approval.ApprovalTemplate`, the ``escalate``
        and ``deescalate`` actors will be the same as the ``approve`` actor, because those events
        are automatically triggered by the ``approve`` event.

        Args:
            event_name (str): The name of the :class:`~sym.sdk.event.Event` to retrieve the
                actor for.

        Returns:
            The :class:`~sym.sdk.user.User` who triggered the :class:`~sym.sdk.event.Event`
            with the given ``event_name``, if that :class:`~sym.sdk.event.Event` has occurred
            in the current :class:`~sym.sdk.flow.Run`. Otherwise returns ``None``.
        """

    @abstractmethod
    def get_context(self, event_name: str) -> Optional[Dict[str, Any]]:
        """Retrieve a dict containing the context values attached to a specific
        :class:`~sym.sdk.event.Event`.

        For example, for a :class:`~sym.sdk.flow.Run` using the
        :class:`~sym.sdk.templates.approval.ApprovalTemplate`, to get the context attached to the
        initial request :class:`~sym.sdk.event.Event`::

            context = event.get_context("request")

        For a list of event names, see the relevant template's Enum (e.g.
        :class:`~sym.sdk.templates.approval.ApprovalTemplateStep`).

        Args:
            event_name (str): The name of the :class:`~sym.sdk.event.Event` to retrieve the
                actor for.

        Returns:
            A dict representing the context attached to the :class:`~sym.sdk.event.Event`
            with the given ``event_name``, if that :class:`~sym.sdk.event.Event` has occurred
            in the current :class:`~sym.sdk.flow.Run`. Otherwise returns ``None``.
        """



from typing import Callable, List

from sym.sdk.event import Event
from sym.sdk.field_option import FieldOption


def reducer(fn):
    """A decorator marking this method as a Reducer.

    Reducers take in an Event and return a single value.py

    Reducer names are always prefixed with ``get_``."""
    return fn


def hook(fn):
    """A decorator marking this method as a Hook.

    Hooks allow you to alter control flow by overriding default implementations of Template steps.
    Hook names may be prefixed with ``on_`` or ``after_``.

    If a Hook name is prefixed with ``on_``, it will be executed after the :class:`~sym.sdk.events.Event` is fired
    but before the default implementation.

    If a Hook name is prefixed with ``after_``, it will be executed after the default implementation.
    """
    return fn


def prefetch(field_name: str) -> Callable[[Event], List[FieldOption]]:
    """A decorator marking this method as a Prefetch Reducer.

    Prefetch Reducers allow you to retrieve a list of options for the Prompt Field identified by `field_name`.
    The logic will be executed prior to displaying the request form to users in Slack, such that the options
    are dynamically populated with the values returned by this Reducer.

    Note: Prefetch Reducers are time-sensitive! The request form will display "Loading..." until all Prefetch Reducers have been executed, so long-running Prefetch Reducers may result in performance delays.
    """



"""Representations of Users in both Sym and third parties."""

import itertools
from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict, List, Optional
from uuid import UUID

from .resource import SymBaseResource


class UserIdentity(ABC, SymBaseResource):
    """Represents a :class:`~sym.sdk.user.User`'s
    identity in an external system such as Slack or PagerDuty.
    """

    def __str__(self) -> str:
        return f"Identity({self.service}: {self.user_id})"

    def __repr__(self) -> str:
        return str(self)

    @property
    @abstractmethod
    def service(self) -> str:
        """The name of the external system providing the identity.

        For example, :mod:`~sym.sdk.integrations.slack`.
        """

    @property
    @abstractmethod
    def service_id(self) -> str:
        """The ID of the external system providing the identity.

        For example, "T123ABC" for a Slack Workspace.
        """

    @property
    @abstractmethod
    def user_id(self) -> str:
        """The :class:`~sym.sdk.user.User`'s identifier in the external system.

        For example, the :class:`~sym.sdk.user.User`'s Slack ID.
        """


class UserRole(str, Enum):
    """UserRoles represent very basic permission sets for users in the Sym platform.

    UserRoles are defined in descending order of permissions. Each role should include
    all permissions of the roles below it.
    """

    ADMIN = "admin"
    """Admins are allowed to do anything.
    (e.g. Use symflow CLI, run Flows, apply Terraform)
    """

    MEMBER = "member"
    """Members can run Flows and interact with requests."""

    GUEST = "guest"
    """Guests can approve/deny/revoke requests, if the Flow is configured to allow guest interactions"""


class User(ABC, SymBaseResource):
    """The atomic representation of a user in Sym.

    :class:`~sym.sdk.user.User`s have many :class:`~sym.sdk.user.UserIdentity`,
    which are used for referencing said user in external systems.
    """

    def __str__(self) -> str:
        identities = list(itertools.chain.from_iterable(self.identities.values()))
        return f"User(email={self.email}, username={self.username}, identities={identities})"

    def __repr__(self) -> str:
        return f"User({self.username})"

    def __hash__(self):
        return hash(self.__class__) + hash(self.username)

    def __eq__(self, other):
        if not isinstance(other, User):
            return False
        return self.username == other.username

    @property
    @abstractmethod
    def id(self) -> UUID:
        """The :class:`~sym.sdk.user.User`'s globally unique identifier."""

    @property
    @abstractmethod
    def email(self) -> Optional[str]:
        """The :class:`~sym.sdk.user.User`'s email if the user is of type "normal", or None otherwise."""

    @property
    @abstractmethod
    def username(self) -> str:
        """The :class:`~sym.sdk.user.User`'s username if the user is of type "bot", or
        the email if the user is of type "normal".
        """

    @property
    @abstractmethod
    def type(self) -> str:
        """The :class:`~sym.sdk.user.User`'s type (i.e., "bot" or "normal")."""

    @property
    @abstractmethod
    def first_name(self) -> Optional[str]:
        """The :class:`~sym.sdk.user.User`'s first name."""

    @property
    @abstractmethod
    def last_name(self) -> Optional[str]:
        """The :class:`~sym.sdk.user.User`'s last name."""

    @property
    @abstractmethod
    def identities(self) -> Dict[str, List[UserIdentity]]:
        """Retrieves the set of identities associated with this
        :class:`~sym.sdk.user.User`, grouped by service type.

        A mapping of service types to lists of :class:`~sym.sdk.user.UserIdentity`.
        """

    @property
    @abstractmethod
    def role(self) -> UserRole:
        """The :class:`~sym.sdk.user.User`'s role (i.e., "admin", "member", or "guest")."""

    @abstractmethod
    def identity(
        self,
        service_type: str,
        service_id: Optional[str] = None,
    ) -> Optional[UserIdentity]:
        """Retrieves this :class:`~sym.sdk.user.User`'s :class:`~sym.sdk.user.UserIdentity`
        for a particular external system.

        External systems specified by a service_type, and optionally a service_id.

        Args:
            service_type: The name of one of Sym's :mod:`~sym.sdk.integrations`.
            service_id: An identifier for an instance of a service, such as a Slack Workspace ID.

        Returns:
            A :class:`~sym.sdk.user.UserIdentity`, or None if no identity is found for the Integration.
        """



import pytest as pytest

from sym.sdk import SlackChannelID, SlackChannelName, SlackUser, SlackUserGroup


class TestRequestDestination:
    def test_slack_channel_id_succeeds_for_valid_channel_ids(self):
        assert SlackChannelID(channel_id="C12345").channel_id == "C12345"
        assert SlackChannelID(channel_id="D12345").channel_id == "D12345"
        assert SlackChannelID(channel_id="G12345").channel_id == "G12345"
        assert SlackChannelID(channel_id="  C12345 ").channel_id == "C12345"

    @pytest.mark.parametrize(
        "channel_id, match_error",
        [
            ("", "must be non-empty"),
            (" ", "must be non-empty"),
            ("X123", "not a valid Slack Channel ID"),
        ],
    )
    def test_slack_channel_id_validate_channel_id_format_throws_error_if_not_valid_channel_id(
        self, channel_id, match_error
    ):
        with pytest.raises(ValueError, match=match_error):
            SlackChannelID(channel_id=channel_id)

    def test_slack_channel_id__str__(self):
        assert SlackChannelID(channel_id="C12345").__str__() == "SlackChannelID 'C12345'"
        assert SlackChannelID(channel_id="D12345").__str__() == "SlackChannelID 'D12345'"
        assert SlackChannelID(channel_id="G12345").__str__() == "SlackChannelID 'G12345'"
        assert SlackChannelID(channel_id="  C12345 ").__str__() == "SlackChannelID 'C12345'"

    def test_slack_channel_name_succeeds_if_channel_name_exists(self):
        assert (
            SlackChannelName(channel_name="is-valid-channel-name").channel_name
            == "is-valid-channel-name"
        )
        assert (
            SlackChannelName(channel_name="  is-valid-channel-name  ").channel_name
            == "is-valid-channel-name"
        )

    @pytest.mark.parametrize(
        "channel_name, match_error",
        [
            ("", "channel_name must be non-empty"),
            (" ", "channel_name must be non-empty"),
        ],
    )
    def test_slack_channel_name_throws_error_when_no_value_given(self, channel_name, match_error):
        with pytest.raises(ValueError, match=match_error):
            SlackChannelName(channel_name=channel_name)

    def test_slack_channel_name__str__(self):
        assert (
            SlackChannelName(channel_name="is-valid-channel-name").__str__()
            == "SlackChannelName 'is-valid-channel-name'"
        )
        assert (
            SlackChannelName(channel_name="  is-valid-channel-name  ").__str__()
            == "SlackChannelName 'is-valid-channel-name'"
        )

    def test_slack_user_succeeds_if_valid(self):
        assert SlackUser(user_id="W09876").user_id == "W09876"
        assert SlackUser(user_id="U12345").user_id == "U12345"
        assert SlackUser(user_id="  U12345  ").user_id == "U12345"

    @pytest.mark.parametrize(
        "user_id, match_error",
        [
            ("", "user_id must be non-empty"),
            (" ", "user_id must be non-empty"),
            (
                "X123",
                "X123 is not a valid Slack User ID. Slack User IDs must start with 'U' or 'W'.",
            ),
        ],
    )
    def test_slack_user_throws_error_if_invalid_user_id(self, user_id, match_error):
        with pytest.raises(ValueError, match=match_error):
            SlackUser(user_id=user_id)

    def test_slack_user__str__(self):
        assert SlackUser(user_id="U12345").__str__() == "SlackUser 'U12345'"
        assert SlackUser(user_id="  U12345  ").__str__() == "SlackUser 'U12345'"

    def test_slack_user_mention(self):
        assert SlackUser(user_id="U12345").mention() == "<@U12345>"
        assert SlackUser(user_id="  U12345  ").mention() == "<@U12345>"

    def test_slack_user_group_if_valid(self):
        assert SlackUserGroup(users=[SlackUser(user_id="U12345")]).users == [
            SlackUser(user_id="U12345")
        ]

    @pytest.mark.parametrize(
        "users, match_error",
        [
            ([], "at least 1 items"),
            (
                [
                    SlackUser(user_id="U1"),
                    SlackUser(user_id="U2"),
                    SlackUser(user_id="U3"),
                    SlackUser(user_id="U4"),
                    SlackUser(user_id="U5"),
                    SlackUser(user_id="U6"),
                    SlackUser(user_id="U7"),
                    SlackUser(user_id="U8"),
                ],
                "at most 7 items",
            ),
        ],
    )
    def test_slack_user_group_throws_error_if_invalid(self, users, match_error):
        with pytest.raises(ValueError, match=match_error):
            SlackUserGroup(users=users)

    def test_slack_user_group__str__(self):
        assert (
            SlackUserGroup(
                users=[SlackUser(user_id="U12345"), SlackUser(user_id="U6789")]
            ).__str__()
            == "SlackUserGroup '['U12345', 'U6789']'"
        )



"""Helpers for interacting with the OneLogin API within the Sym SDK."""
from typing import List, Optional

from sym.sdk.user import User


def list_roles(name_filter: Optional[str] = None) -> List[dict]:
    """Returns a list of all roles in the OneLogin account configured in the current
    :class:`~sym.sdk.flow.Environment`. Roles are represented as dictionaries, whose
    structure is defined by the OneLogin API.

    Details here: https://developers.onelogin.com/api-docs/2/roles/list-roles

    Note that if more than 650 roles exist in the OneLogin account, this call will
    fail. This is a limitation of the OneLogin API.

    Args:
        name_filter: If provided, only roles whose name contains the given string will
            be returned.

    Returns:
        A list of dictionaries representing OneLogin roles.
    """


def get_user_info(user: User) -> dict:
    """Get information about a OneLogin user.

    Refer to OneLogin's get-user API documentation for details on the response format:
    https://developers.onelogin.com/api-docs/2/users/get-user

    Args:
        user: The Sym :class:`~sym.sdk.user.User` to request OneLogin info for

    Returns:
        A dictionary of user information
    """


def users_in_role(role_id: int) -> List[User]:
    """Returns the members of the specified OneLogin role.

    The OneLogin role's ID must be given, and the function will check that the role
    exists and is accessible.

    Args:
        role_id: The ID of the OneLogin role.

    Returns:
        A list of :class:`Users <sym.sdk.user.User>` who are members of the given role.

    Raises:
        :class:`~sym.sdk.exceptions.sdk.MissingArgument`: If the role_id is not provided.
        :class:`~sym.sdk.exceptions.onelogin.OneLoginError`: If the role does not exist or is not accessible.
    """


def is_user_in_role(user: User, *, role_id: int) -> bool:
    """Checks if the provided :class:`~sym.sdk.user.User` is a member of the OneLogin role specified.

    The OneLogin role's ID must be given, and the function will check that the role
    exists and is accessible.

    Args:
        user: The :class:`~sym.sdk.user.User` to check role membership for.
        role_id: The ID of the OneLogin role.

    Returns:
        True if the user is a member of the specified OneLogin role, False otherwise.

    Raises:
        :class:`~sym.sdk.exceptions.sdk.MissingArgument`: If the role_id is not provided.
        :class:`~sym.sdk.exceptions.onelogin.OneLoginError`: If the role does not exist or is not accessible,
            or if no :class:`~sym.sdk.user.User` is provided.
    """



from sym.sdk.exceptions.sym_exception import SymException


class SDKError(SymException):
    """Generic exceptions raised by the Sym SDK at runtime.

    Args:
        name: The name of the exception (used as the second part of the error_code, e.g. MISSING_ARGUMENT)
        message: The exception message to display
    """

    def __init__(self, name: str, message: str, error_type: str = "SymSDK"):
        super().__init__(error_type=error_type, name=name, message=message)


class MissingArgument(SDKError):
    """This error is raised when a Sym SDK method is called without the required arguments.

    Args:
        arg: The name of the missing argument
        method: The name of the SDk method that was called
    """

    def __init__(self, arg: str, method: str):
        super().__init__(
            name="MISSING_ARGUMENT",
            message=(
                f"Required argument '{arg}' for method '{method}' not provided. "
                "Please check your implementation file to ensure that all required arguments are passed."
            ),
        )



import sym

__all__ = [
    "AccessTarget",
    "ApprovalTemplate",
    "Channel",
    "Event",
    "EventMeta",
    "FieldOption",
    "Flow",
    "Payload",
    "RequestDestination",
    "RequestDestinationFallback",
    "Run",
    "SRN",
    "SlackChannelID",
    "SlackChannelName",
    "SlackUser",
    "SlackUserGroup",
    "SymIntegrationError",
    "SymResource",
    "SymSDKError",
    "Template",
    "User",
    "UserIdentity",
    "UserRole",
    "aws_lambda",
    "github",
    "hook",
    "okta",
    "onelogin",
    "pagerduty",
    "prefetch",
    "reducer",
    "slack",
]

if getattr(sym, "initialized", True):
    # The Sym Runtime requires delayed initialization to prevent circular dependencies.
    from .annotations import hook, prefetch, reducer
    from .errors import SymIntegrationError, SymSDKError
    from .event import Channel
    from .flow import Flow, Run
    from .integrations import aws_lambda, github, okta, onelogin, pagerduty, slack
    from .models import *  # noqa: F403
    from .templates import ApprovalTemplate, Template



from sym.sdk.exceptions.sym_exception import SymException


class AWSError(SymException):
    """This is the base class for all AWSErrors

    Args:
        name: The name of the exception (used as the second part of the error_code, e.g. ASSUME_ROLE_FAILED)
        message: The exception message to display"""

    def __init__(self, name: str, message: str, error_type: str = "AWS"):
        super().__init__(error_type=error_type, name=name, message=message)


class AWSLambdaError(AWSError):
    """This is the base class for all AWS Lambda exceptions raised by the Sym Runtime.

    Args:
        name: The name of the exception (used as the second part of the error_code, e.g. FATAL_FUNCTION_ERROR)
        message: The exception message to display
    """

    def __init__(self, name: str, message: str):
        super().__init__(error_type="AWSLambda", name=name, message=message)


class AWSIAMError(AWSError):
    """This is the base class for all AWS IAM exceptions raised by the Sym Runtime.

    Args:
        name: The name of the exception (used as the second part of the error_code, e.g. GROUP_NOT_FOUND)
        message: The exception message to display
    """

    def __init__(self, name: str, message: str):
        super().__init__(error_type="AWSIAM", name=name, message=message)


class AWSSSOError(AWSError):
    """This is the base class for all AWS SSO exceptions raised by the Sym Runtime.

    Args:
        name: The name of the exception (used as the second part of the error_code, e.g. GROUP_NOT_FOUND)
        message: The exception message to display
    """

    def __init__(self, name: str, message: str):
        super().__init__(error_type="AWSSSO", name=name, message=message)



"""Integrations with a host of third-party services."""

__all__ = ["aws_iam", "aws_lambda", "aws_sso", "github", "okta", "onelogin", "pagerduty", "slack"]

from .aws_iam import *
from .aws_lambda import *
from .aws_sso import *
from .github import *
from .okta import *
from .onelogin import *
from .pagerduty import *
from .slack import *



from abc import ABC, abstractmethod
from typing import Optional, final

from sym.sdk.event import Event
from sym.sdk.exceptions.identity import IdentityNotFound
from sym.sdk.user import User
from sym.sdk.utils import persist_user_identity

from .integration import Integration


class AccessStrategy(ABC):
    """A standardized interface that allows Implementers to define custom Strategies, enabling Sym
    to perform escalations and de-escalations against almost any Internet-accessible service."""

    @final
    def __init__(self, integration: Integration):
        self.integration = integration

    @final
    def get_requester_identity(self, event: Event) -> str:
        """Retrieves the external identity of the user who initiated the request represented by the
        ``event`` parameter; first from stored user information, then by calling
        :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.fetch_remote_identity` as needed.

        This method is meant to be called in implementations of the ``AccessStrategy`` class; e.g.,
        :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.escalate` and
        :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.deescalate`, as a way to resolve
        the mapping of a Sym user to their corresponding identity in the external service being
        interacted with.

        To be clear, this method does not *need* to be called in situations where there is no need
        to maintain a mapping of users to external identities; for example, if the user's email
        address, which is already captured by Sym, is all that is needed by the external service.

        This method's source can also be used as a reference for how the external identity lookup
        process works. It should not be overridden!

        Args:
            event: An object containing information about the current event

        Returns:
            The unique identifier for the user who initiated the request, for the external service
            being interacted with.

        Raises:
            IdentityNotFound: If the user has no identity mapping configured and it could not be
                looked up (``fetch_remote_identity()`` returned ``None``)
        """
        requester: User = event.get_actor("request")
        if requester_id := requester.identity(self.integration.type, self.integration.external_id):
            return requester_id.user_id

        if not (requester_id := self.fetch_remote_identity(requester)):
            raise IdentityNotFound(email=requester.email, service_id=self.integration.external_id)

        persist_user_identity(
            email=requester.email,
            service=self.integration.type,
            service_id=self.integration.external_id,
            user_id=requester_id,
        )
        return requester_id

    @abstractmethod
    def fetch_remote_identity(self, user: User) -> Optional[str]:
        """Fetches the external identity of the requester from the remote service; not meant to be
        called directly.

        This method is used to fetch the external identity of the requester when Sym does not
        already have that information available. Commonly, this is when a user makes a request
        utilizing a given :class:`~sym.sdk.strategies.integration.Integration` for the first time.

        This method should return a string that represents the user's unique user ID in the
        service this Strategy interacts with; or, more specifically, whatever is needed to identify
        the user being escalated or de-escalated in the service that is performing the escalation/
        de-escalation. The user ID value can be any string; it is treated as an opaque value by the
        Sym platform. For example, if the Strategy were interacting with AWS IAM, the user ID
        returned by this method would be the user's ARN.

        If the user's identity cannot be fetched, or this automated remote identity discovery does
        not need to be implemented, this method may return ``None`` instead; however, note that
        ``get_requester_identity()`` will raise an
        :class:`~sym.sdk.exceptions.identity.IdentityNotFound` exception if the requester does not
        already have an identity.

        To be clear, this method does not *need* to be implemented in situations where there is no
        need to maintain a mapping of users to external identities; for example, if the user's email
        address, which is already captured by Sym, is all that is needed by the external service.

        Implementers should not call this method directly; instead, use
        :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.get_requester_identity`,
        which calls this method as needed.

        If this method returns a non-``None`` value, ``get_requester_identity()`` will persist the
        value to the user saved in Sym so that this lookup doesn't need to be performed again.

        Args:
            user: The user to fetch an external service identity for

        Returns:
            An opaque string value representing the user's unique identifier in the external
            service, or ``None`` if this lookup should not or cannot be performed.
        """

    @abstractmethod
    def escalate(self, target_id: str, event: Event) -> Optional[dict]:
        """Perform an escalation in the external service.

        This method is called to perform escalations, and should implement all functionality needed
        to grant access to a user for a resource. For example, an implementation of this method
        might construct a JSON request body and then submit an HTTP POST request to some REST API
        endpoint.

        This method should be idempotent; escalating the same user and target pair multiple times
        must not cause an error. However, errors may be reported by simply raising an exception;
        these exceptions will be handled by Sym's built-in error handling logic. If you wish to
        implement your own exceptions, please refer to
        :class:`~sym.sdk.exceptions.sym_exception.SymException` and
        :class:`~sym.sdk.exceptions.sym_exception.ExceptionWithHint`.

        Args:
            target_id: The ID of the "target" of the escalation, which can be thought of what the
                user is being escalated to. Specified by the Implementer in Terraform, and often
                selected by the requesting user when making the request.
            event: A representation of the escalation event. The ID of the user being escalated can
                be retrieved by calling
                :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.get_requester_identity`
                on ``event``. It also contains other useful values; see the docs for
                :class:`~sym.sdk.event.Event`.

        Returns:
            An optional dictionary of state that should be preserved for de-escalation. This can
            be accessed in :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.deescalate` by
            calling :func:`~sym.sdk.templates.step.get_step_output` in the implementation there.
            This value is also made available to `Hooks <https://docs.symops.com/docs/hooks>`_ in
            Flow implementations, also via ``get_step_output()``. This method can also return an
            empty dict or ``None`` to indicate that no special state needs to be preserved for de-
            escalation or Hooks. **Failure must be indicated by raising an exception.**
        """

    @abstractmethod
    def deescalate(self, target_id: str, event: Event) -> Optional[dict]:
        """Perform a de-escalation in the external service.

        This method is called to perform de-escalations, and should implement all functionality
        needed to revoke access granted by this strategy's
        :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.escalate` method. For example, an
        implementation of this method might construct a JSON request body and then submit an HTTP
        DELETE request to some REST API endpoint.

        This method should be idempotent; de-escalating the same user and target pair multiple times
        (or calling this method on a user who is not escalated at the time of the call) must not
        cause an error. However, errors may be reported by simply raising an exception; these
        exceptions will be handled by Sym's built-in error handling logic. If you wish to implement
        your own exceptions, please refer to
        :class:`~sym.sdk.exceptions.sym_exception.SymException` and
        :class:`~sym.sdk.exceptions.sym_exception.ExceptionWithHint`.

        Args:
            target_id: The ID of the "target" of the de-escalation, which can be thought of what the
                user is being removed from for de-escalation. Specified by the Implementer in
                Terraform, and often was selected by the requesting user when making the request;
                will be the same value as was passed to
                :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.escalate`
            event: A representation of the de-escalation event. The ID of the user being
                de-escalated can be retrieved by calling
                :func:`~sym.sdk.strategies.access_strategy.AccessStrategy.get_requester_identity`
                on ``event``. It also contains other useful values; see the docs for Event.

        Returns:
            An optional dictionary of state that should be made available to
            `Hooks <https://docs.symops.com/docs/hooks>`_ in Flow implementations via
            :func:`~sym.sdk.templates.step.get_step_output`. This method can also return an empty
            dict or ``None`` to indicate that nothing needs to be preserved for Hooks. **Failure
            must be indicated by raising an exception.**
        """



from .event import Channel, Event, EventMeta, Payload  # noqa
from .field_option import FieldOption  # noqa
from .flow import Environment, Flow, Run  # noqa
from .request_destination import (  # noqa
    RequestDestination,
    RequestDestinationFallback,
    SlackChannelID,
    SlackChannelName,
    SlackUser,
    SlackUserGroup,
)
from .resource import SRN, SymResource  # noqa
from .target import AccessTarget  # noqa
from .user import User, UserIdentity, UserRole  # noqa



import pytest

from sym.sdk.flow import Flow, Run


class TestFlow:
    def test_flow_abstract(self):
        with pytest.raises(TypeError, match="Can't instantiate abstract class Flow"):
            Flow()  # type: ignore

    def test_run_abstract(self):
        with pytest.raises(TypeError, match="Can't instantiate abstract class Run"):
            Run()  # type: ignore



from typing import List, Optional

from pydantic import BaseModel, Field, validator

"""Representation of different options for routing Sym Requests."""


class RequestDestination(BaseModel):
    """A super-class for classes representing a Sym Request Destination."""

    allow_self: bool = False
    """A boolean indicating whether or not the requester may approve this Request."""

    timeout: Optional[int] = None
    """An optional integer representing the duration until this destination will time out, in seconds.
    If not specified, then the destination will never time out, i.e the Request will hang indefinitely until responded 
    to.
    """


class SlackChannelID(RequestDestination):
    """A Request to be sent to a Slack Channel, identified by a Slack Channel ID, e.g. 'C12345'."""

    channel_id: str
    """A string that identifies a Slack Channel by its unique ID. e.g. 'C12345'."""

    @validator("channel_id")
    def validate_channel_id_format(cls, value):
        """Validates that the given `channel_id` follows Slack's convention of starting with a C, D or G.
        (e.g. 'D12345').
        """
        value = value.strip()

        if not value:
            raise ValueError("channel_id must be non-empty")

        if not value[0] in {"C", "D", "G"}:
            raise ValueError(
                f"{value} is not a valid Slack Channel ID. Channel IDs must start with one of 'C', 'D', or 'G'."
            )

        return value

    def __str__(self):
        """A method that will print a human readable explanation of the type of data,
        in this case a SlackChannelId with a string of the channel id."""
        return f"SlackChannelID '{self.channel_id}'"


class SlackChannelName(RequestDestination):
    """A Request to be sent to a Slack Channel, identified by a Slack Channel name."""

    channel_name: str
    """A string that identifies a Slack Channel by its unique name. (e.g. '#sym-errors')."""

    def __str__(self):
        """A method that will print a human readable explanation of the type of data,
        in this case a SlackChannelName with a string of the channel name."""
        return f"SlackChannelName '{self.channel_name}'"

    @validator("channel_name")
    def validate_channel_name_format(cls, value):
        value = value.strip()

        if not value:
            raise ValueError("channel_name must be non-empty")

        return value


class SlackUser(RequestDestination):
    """A Request to be sent to a Slack User, identified by a Slack User ID."""

    user_id: str
    """A string that identifies a Slack User by their unique ID (also called Member ID). (e.g. 'U12345')."""

    @validator("user_id")
    def validate_user_id_format(cls, value):
        """Validates that the given `user_id` follows Slack's convention of starting with a U or W. (e.g. 'U12345')."""
        value = value.strip()

        if not value:
            raise ValueError("user_id must be non-empty")

        if not value[0] in {"U", "W"}:
            raise ValueError(
                f"{value} is not a valid Slack User ID. Slack User IDs must start with 'U' or 'W'."
            )

        return value

    def mention(self) -> str:
        """Returns a string that can be used to @mention this User in a Slack message."""
        return f"<@{self.user_id}>"

    def __str__(self):
        """Returns a human readable string indicating the type of data, in this case Slack User and the user_id"""
        return f"SlackUser '{self.user_id}'"


class SlackUserGroup(RequestDestination):
    """A Request to be sent in a Slack Group DM including the given Slack Users, with a maximum of 7 Users."""

    users: List[SlackUser] = Field(..., min_items=1, max_items=7)
    """A list of :class:`~sym.sdk.request_destination.SlackUser` objects to include in the group DM.
    Note: While each SlackUser object supports an `allow_self` attribute, this attribute will ignored in favor of the
    `allow_self` attribute set on the SlackUserGroup object.
    """

    def __str__(self):
        """Returns a human readable string indicating the type of data, in this case SlackUserGroup
        and the user IDs printed in a comma seperated list"""
        user_ids = [user.user_id for user in self.users]
        return f"SlackUserGroup '{user_ids}'"


class RequestDestinationFallback(BaseModel):
    """A list of :class:`~sym.sdk.request_destination.RequestDestination` to attempt to send requests to.
    The next :class:`~sym.sdk.request_destination.RequestDestination` will be attempted based on the failure mode
    configuration.
    """

    destinations: List[RequestDestination]
    """The list of :class:`~sym.sdk.request_destination.RequestDestination` objects to attempt to send requests to."""

    continue_on_delivery_failure: bool = True
    """If set to True, if Sym fails to deliver a Request to a :class:`~sym.sdk.request_destination.RequestDestination`
    (such as when the Slack Channel does not exist), then the next
    :class:`~sym.sdk.request_destination.RequestDestination` in the `destinations` list will be attempted."""

    continue_on_timeout: bool = False
    """If set to True, if the current Request times out, then a new Request will be sent to the next
    :class:`~sym.sdk.request_destination.RequestDestination` in the `destinations` list."""


class RequestForwardContext(BaseModel):
    """An object containing additional context about the current and next destinations of the Sym Request. An instance
    of this class is included as an argument when `on_request_forward` and `after_request_forward` hooks are invoked by
    the Sym Runtime.
    """

    all_request_destinations: List[RequestDestination]
    """A list of all :class:`~sym.sdk.request_destination.RequestDestination` instances that this Sym Request may be 
    sent to"""
    current_destination_index: int
    """An integer representing the index of the current destination in `all_request_destinations`"""
    next_destination_index: Optional[int]
    """An integer representing the index of the next destination in `all_request_destinations`. When set to `None`, 
    this means that the `current_destination_index` is at the last element of `all_request_destinations`, i.e there 
    are no more destinations to try next.
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sym',
 'sym.sdk',
 'sym.sdk.annotations',
 'sym.sdk.exceptions',
 'sym.sdk.integrations',
 'sym.sdk.strategies',
 'sym.sdk.templates',
 'sym.sdk.tests',
 'sym.sdk.tests.annotations',
 'sym.sdk.tests.integrations',
 'sym.sdk.tests.static',
 'sym.sdk.utils']

package_data = \
{'': ['*']}

install_requires = \
['bcrypt>=3.2,<4.0',
 'pydantic>=1.7.3,<2.0.0',
 'requests>=2.31.0,<3.0.0',
 'structlog==21.5.0',
 'virtualenv>=20.4.7,<21.0.0']

setup_kwargs = {
    'name': 'sym-sdk',
    'version': '0.39.0',
    'description': "Sym's Python SDK",
    'long_description': 'Sym Python SDK\n================\n\n`Sym <https://symops.com/>`_ is the security workflow platform made for engineers, by engineers.\n\nWe solve the intent-to-execution gap between policies and workflows by providing fast-moving engineering teams with the just-right primitives to roll out best-practice controls.\n\nThis is the Python SDK for Sym.\nFor guides and other help, check out our `main docs site <https://docs.symops.com/>`_.\n\nThe SDK docs are broken into several core modules, which are described below.\nClick on one to see the classes and functions available in your `Handlers <https://docs.symops.com/docs/workflow-handlers>`_.\n\nThe Sym SDK is used to customize workflow templates that are exposed by our `Terraform provider <https://registry.terraform.io/providers/symopsio/sym/latest/docs>`_. Here\'s an example using the ``sym:approve`` Template!\n\n.. code-block:: python\n\n   from sym.sdk.annotations import reducer\n   from sym.sdk.integrations import pagerduty, okta, slack\n\n   @reducer\n   def get_approvers(evt):\n      # The import here uses credentials defined in an Integration in Terraform\n      if pagerduty.is_on_call(evt.user, schedule="id_of_eng_on_call"):\n         # This is a self-approval in a DM\n         return slack.user(evt.user)\n\n      if evt.payload.fields["urgency"] == "Emergency":\n         # This is a self-approval in a channel\n         return slack.channel("#break-glass", allow_self=True)\n\n      on_call_mgrs = okta.group("OnCallManagers").members()\n      # This would cause each on-call manager to be DMed\n      return slack.group([slack.user(x) for x in on_call_mgrs])\n\nIf you\'re interested in using Sym, please `reach out <https://symops.com/sales>`_!\n',
    'author': 'Sym Engineering',
    'author_email': 'pypi@symops.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://sdk.docs.symops.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.6,<4.0.0',
}


setup(**setup_kwargs)



###
### Begin pure Slack samples
###

SIMPLE_SLACK_CHANNEL_CODE = """
from sym.sdk.annotations import reducer
from sym.sdk.integrations import slack

@reducer
def get_approvers(event):
    return slack.channel('#managers')
"""

SIMPLE_SLACK_USER_CODE = """
from sym.sdk.annotations import reducer
from sym.sdk.integrations import slack

@reducer
def get_approvers(event):
    return slack.user('@David Ruiz')
"""

SIMPLE_SLACK_USER_ID_CODE = """
from sym.sdk.annotations import reducer
from sym.sdk.integrations import slack

@reducer
def get_approvers(event):
    return slack.user(event.payload.user)
"""

SIMPLE_SLACK_GROUP_CODE = """
from sym.sdk.annotations import reducer
from sym.sdk.integrations import slack

@reducer
def get_approvers(event):
    return slack.group(['@David Ruiz', '@Jon Demo'])
"""

SLACK_GROUP_WITH_USERS_CODE = """
from sym.sdk.annotations import reducer
from sym.sdk.integrations import slack, pagerduty

@reducer
def get_approvers(event):
    return slack.group([event.user, event.user])
"""

SIMPLE_ALLOW_SELF_FALSE_CODE = """
import datetime
from sym.sdk.annotations import reducer
from sym.sdk.integrations import slack
@reducer
def get_approvers(event):
    return slack.channel("#break-glass", allow_self=False)
"""

SIMPLE_ALLOW_SELF_TRUE_CODE = """
import datetime
from sym.sdk.annotations import reducer
from sym.sdk.integrations import slack
@reducer
def get_approvers(event):
    return slack.channel("#break-glass", allow_self=True)
"""

GET_APPROVERS_USING_EVENT_CODE = """
import datetime
from sym.sdk.annotations import reducer
from sym.sdk.integrations import slack
@reducer
def get_approvers(event):
    if event.fields["Urgency"] == "High":
        return slack.channel("#break-glass")

    if datetime.datetime.now().hour >= 17:
        return slack.user('@Ari Tang')
    else:
        return [slack.user(u) for u in ['@David Ruiz', '@Jon Demo']]
"""

###
### Begin PagerDuty samples
###

SIMPLE_PAGERDUTY_IS_ON_CALL_CODE = """
from sym.sdk.annotations import reducer
from sym.sdk.integrations import pagerduty, slack

@reducer
def get_approvers(event):
    if pagerduty.is_on_call(event.user):
        return slack.channel("#on-call")
    return slack.channel("#not-on-call")
"""

SIMPLE_PAGERDUTY_IS_ON_CALL_SCHEDULE_CODE = """
from sym.sdk.annotations import reducer
from sym.sdk.integrations import pagerduty, slack

@reducer
def get_approvers(event):
    if pagerduty.is_on_call(event.user, schedule_name="test-schedule"):
        return slack.channel("#on-call")
    return slack.channel("#not-on-call")
"""

SIMPLE_PAGERDUTY_USERS_ON_CALL_CODE = """
from sym.sdk.annotations import reducer
from sym.sdk.integrations import pagerduty, slack

@reducer
def get_approvers(event):
    return [slack.user(u) for u in pagerduty.users_on_call()]
"""

SIMPLE_PAGERDUTY_USERS_ON_CALL_SCHEDULE_CODE = """
from sym.sdk.annotations import reducer
from sym.sdk.integrations import pagerduty, slack

@reducer
def get_approvers(event):
    return [slack.user(u) for u in pagerduty.users_on_call(schedule_name="test-schedule")]
"""



from abc import ABC, abstractmethod
from typing import Any, Dict

from sym.sdk.resource import SymResource


class Integration(ABC, SymResource):
    """Represents an Integration (service-specific config and state) in the Sym SDK.

    For the purposes of the SDK, Integrations can be thought of as the "state" of a Strategy. Or, to
    put it another way, if a Strategy tells Sym how to interact with an external service to perform
    escalations, the Integration is the service-specific state Sym maintains, including the identity
    mappings of Sym users to the external service."""

    @property
    @abstractmethod
    def type(self) -> str:
        """The type of service represented by this Integration object."""

    @property
    @abstractmethod
    def external_id(self) -> str:
        """A unique (for this type), meaningful identifier for this Integration.

        This allows you to have multiple integrations of the same type. The meaning of the
        ``external_id`` varies from Strategy to Strategy—some Strategies use this as a domain or
        tenant identifier for some services, while for others it's simply an Implementer-defined
        identifier.

        For example, a Slack Integration uses the Slack workspace ID as the ``external_id`` value,
        while an Okta Integration uses the Okta Domain as the ``external_id``.
        """

    @property
    @abstractmethod
    def settings(self) -> Dict[str, Any]:
        """A dictionary of settings for the Integration, as specified in Terraform. May be empty,
        but will never be ``None``."""



"""Methods for invoking AWS Lambda functions."""

from typing import Any, Optional, Union

from sym.sdk.event import Event
from sym.sdk.exceptions.aws import AWSLambdaError  # noqa


def invoke(arn: str, payload: Optional[Union[Event, dict]] = None) -> Any:
    """Synchronously invokes an AWS Lambda function.

    Args:
        arn: The ARN of the Lambda function to invoke.
        payload: A dict of JSON-serializable data to pass to the function,
            or an :class:`~sym.sdk.event.Event` object.

    Returns:
        The JSON-deserialized object returned by the lambda.
    """


def invoke_async(arn: str, payload: Optional[Union[Event, dict]] = None) -> bool:
    """Asynchronously invokes an AWS Lambda function.

    Args:
        arn: The ARN of the Lambda function to invoke.
        payload: A dict of JSON-serializable data to pass to the function,
            or an :class:`~sym.sdk.event.Event` object.

    Returns:
        A boolean indicating success enqueuing the invocation.
    """



"""Exceptions that can be raised by the Sym Runtime."""

__all__ = [
    "AccessStrategyError",
    "AptibleError",
    "AWSError",
    "AWSIAMError",
    "AWSSSOError",
    "AWSLambdaError",
    "BoundaryError",
    "CouldNotSaveError",
    "ExceptionWithHint",
    "GitHubError",
    "HTTPError",
    "IdentityError",
    "MissingArgument",
    "OktaError",
    "OneLoginError",
    "PagerDutyError",
    "SDKError",
    "SlackError",
    "SymException",
]

from .access_strategy import AccessStrategyError
from .aptible import AptibleError
from .aws import AWSError, AWSIAMError, AWSLambdaError, AWSSSOError
from .boundary import BoundaryError
from .github import GitHubError
from .http import HTTPError
from .identity import CouldNotSaveError, IdentityError
from .okta import OktaError
from .onelogin import OneLoginError
from .pagerduty import PagerDutyError
from .sdk import MissingArgument, SDKError
from .slack import SlackError
from .sym_exception import ExceptionWithHint, SymException



from sym.sdk import hook, reducer


@hook
def simple_hook():
    return "simple_hook"


@reducer
def simple_reducer():
    return "simple_reducer"


class TestReducer:
    def test_all(self):
        assert simple_hook() == "simple_hook"
        assert simple_reducer() == "simple_reducer"
